import logging
import os

import numpy as np
from matplotlib import rc_context

from mrestimator import utility as ut
log = ut.log
from mrestimator.coefficients import CoefficientResult, coefficients
from mrestimator.input_output import *
from mrestimator.fit import *

def full_analysis(
    data=None,
    dt=None,
    kmax=None,
    dtunit=' time unit',
    fitfuncs=None,
    coefficientmethod=None,
    tmin=None,                      # include somehow into 'missing' req. arg
    tmax=None,
    steps=None,                     # dt conversion? optional replace tmin/tmax
    substracttrialaverage=False,
    targetdir=None,
    title=None,                     # overwrites old files in same targetdir
    numboot='auto',                 # optional. default depends on fitfunc
    seed=1,                         # default: 1 uses hard coded seeds
    loglevel=None,                  # only concerns local logfile
    targetplot=None,
    showoverview=True,
    saveoverview=False,
    method=None,                    # overload for coefficientmethod
    ):
    """
        Wrapper function that performs the following four steps:

        - check `data` with `input_handler()`
        - calculate correlation coefficients via `coefficients()`
        - fit autocorrelation function with `fit()`
        - export/plot using the `OutputHandler`

        Usually it should suffice to tweak the arguments and call this
        wrapper function (multiple times).
        Calling the underlying functions individually
        gives slightly more control, though.
        We recommend to set `showoverview=False` when calling in loops to avoid
        opening many figures (and wasting RAM).

        Parameters
        ----------
        data: str, list or numpy.ndarray
            Passed to `input_handler()`. Ideally, import and check data first.
            A `string` is assumed to be the path
            to file(s) that is then imported as pickle or plain text.
            Alternatively, you can provide a `list` or `ndarray` containing
            strings or already imported data. In the latter case,
            `input_handler()` attempts to convert it to the right format.

        dt: float
            How many `dtunits` separate the measurements of the provided data.
            For example, if measurements are taken every 4ms:
            `dt=4`, `dtunit=\'ms\'`.

        kmax: int
            Maximum time lag k (in time steps of size `dt`) to use for
            coefficients. Alternatively, `tmax` or `steps` can be specified

        Other Parameters
        ----------------

        dtunit: str, optional
            Unit description/name of the time steps of the provided data.

        fitfuncs: list, optional
            Which fitfunctions to use e.g. ``fitfuncs=['e', 'eo', 'c']``.
            Renamed from `fitfunctions` in v0.1.4.

        coefficientmethod: str, optional
            `ts` or `sm`, method used for determining the correlation
            coefficients. See the :func:`coefficients` function for details.
            Default is `ts`.

        method: str, optional
            same as `coefficientmethod`, introduced in v0.1.6.

        tmin: float
            Smallest time separation to use for coefficients, in units of
            `dtunit`.
            Only one argument is possible, either `kmax` or `steps` or
            `tmin` and `tmax`.

        tmax: float
            Maximum time separation to use for coefficients.
            For example, to fit the autocorrelation between 8ms and
            2s set: `tmin=8`, `tmax=2000`, `dtunit=\'ms\'`
            (independent of `dt`).

        steps : ~numpy.array, optional
            Specify the fitrange in steps :math:`k` for which to compute
            coefficients :math:`r_k`.
            Note that :math:`k` provided here would need
            to be multiplied with units of [`dt` * `dtunit`] to convert
            back to (real) time.
            If an array of length two is provided, e.g.
            ``steps=(minstep, maxstep)``, all enclosed integer values will be
            used.
            Arrays larger than two are assumed to contain a manual choice of
            steps. Strides other than one are possible.
            Only one argument is possible, either `steps` or `kmax` or
            `tmin` and `tmax`.

        substracttrialaverage: bool, optional
            Substract the average across all trials before calculating
            correlation coefficients.
            Default is `False`.

        targetdir: str, optional
            String containing the path to the target directory where files
            are saved with the filename `title`.
            Per default, `targetdir=None` and no files are written to disk.

        title: str, optional
            String for the filenames. Also sets the main title of the
            overview panel.

        numboot: int or 'auto', optional
            Number of bootstrap samples to draw.
            This repeats every fit `numboot` times so that we can
            provide an uncertainty estimate of the resulting branching
            parameter and autocorrelation time.
            Per default, bootstrapping is only applied in
            `coefficeints()` as most of computing time is needed for the
            fitting. Thereby we have uncertainties on the :math:`r_k`
            (which will be plotted) but each fit is only
            done once.
            Default is `numboot='auto'` where the number of samples depends on
            the fitfunction (100 for the exponential).

        seed : int, None or 'random', optional
            If `numboot` is not zero, provide a seed for the random number
            generator. If `seed=None`, seeding will be skipped.
            Per default, the rng is (re)seeded every time `full_analysis()` is
            called so that every repeated call returns the same error
            estimates.

        showoverview: bool, optional
            Whether to show the overview panel. Default is 'True'.
            Set to 'False' when calling `full_analysis()` repeatedly or just
            saving the panels to disk with `saveoverview` (this temporarily
            overwrites your matplotlib rc parameters for more consistency).
            Otherwise, matplotlib may create large amounts of figures and leak
            memory.
            Note that even when set to 'True' the panel might not show if
            `full_analysis()` is called through a script instead of an
            (interactive) shell, this depends on your matplotlib configuration.

        saveoverview: bool, optional
            Whether to save the overview panel in `targetdir`.
            Default is 'False'.

        loglevel: str, optional
            The loglevel to use for the logfile created
            as `title.log` in the `targetdir`.
            'ERROR', 'WARNING', 'INFO' or 'DEBUG'.
            Per default, no log is written unless `loglevel` and `targetdir`
            are provided.

        targetplot: `matplotlib.axes.Axes`, optional
            You can provide a matplotlib axes element (i.e. a subplot of an
            existing figure) to plot the correlations into.
            The axis will be passed to the `OutputHandler`
            and all plotting will happen within that axes.
            Per default, a new figure is created - that cannot be added
            as a subplot to any other figure later on. This is due to
            the way matplotlib handles subplots.

        Returns
        -------
        OutputHandler
            that is associated
            with the correlation plot, fits and coefficients.
            Also saves meta data and plotted pdfs to `targetdir`.

        Example
        -------

        .. code-block:: python

            # test data, subsampled branching process
            bp = mre.simulate_branching(m=0.95, h=10, subp=0.1, numtrials=50)

            mre.full_analysis(
                data=bp,
                dt=1,
                tmin=0, tmax=1500,
                dtunit='step',
                fitfuncs=['exp', 'exp_offs', 'complex'],
                targetdir='./output',
                title='Branching Process')
        ..
    """

    # ------------------------------------------------------------------ #
    # Arguments
    # ------------------------------------------------------------------ #

    how_to_str = "full_analysis() requires the following arguments:\n" +\
        "'data', 'dt', 'method' and either one of 'kmax', 'tmax' or 'steps'"

    # workaround: if full_analysis() does not reach its end where we remove
    # the local loghandler, it survives and keps logging with the old level
    for hdlr in log.handlers:
        if isinstance(hdlr, logging.FileHandler):
            if hdlr != ut._logfilehandler:
                hdlr.close()
                log.removeHandler(hdlr)

    if data is None or dt is None:
        log.exception(how_to_str)
        raise TypeError

    if kmax is None and tmax is None and steps is None:
        log.exception(how_to_str)
        raise TypeError

    # if there is a targetdir specified, create and use for various output
    if targetdir is not None:
        if isinstance(targetdir, str):
            td = os.path.abspath(os.path.expanduser(targetdir+'/'))
            os.makedirs(td, exist_ok=True)
            ut._set_permissions(td)
            targetdir = td
        else:
            log.exception("Argument 'targetdir' needs to be of type 'str'")
            raise TypeError

        # setup log early so argument errors appear in the logfile
        if loglevel is None:
            # dont create a logfile
            pass
        else:
            if isinstance(loglevel, int) and loglevel > 0:
                pass
            elif str(loglevel).upper() in [
                'ERROR', 'WARNING', 'INFO', 'DEBUG']:
                loglevel = str(loglevel).upper()
            else:
                log.debug(
                    "Unrecognized log level {}, using 'INFO'".format(loglevel))
                loglevel = 'INFO'
            # open new handler and add it to logging module
            loghandler = logging.handlers.RotatingFileHandler(
                targetdir+'/{}.log'.format(
                    'full_analysis' if title is None else title, 'a'),
                maxBytes=5*1024*1024, backupCount=1)
            loghandler.setLevel(logging.getLevelName(loglevel))
            loghandler.setFormatter(ut.CustomExceptionFormatter(
                '%(asctime)s %(levelname)8s: %(message)s',
                "%Y-%m-%d %H:%M:%S"))
            log.addHandler(loghandler)
    else:
        if saveoverview:
            log.warning("Cannot save overview since no targetdir specified, "+\
                "skipping")

    log.debug("full_analysis()")
    if (ut._log_locals):
        log.debug('Locals: {}'.format(locals()))

    try:
        dt = float(dt)
        assert(dt>0)
    except Exception as e:
        log.exception("Argument 'dt' needs to be a float > 0")
        raise

    if not isinstance(dtunit, str):
        log.exception("Argument 'dtunit' needs to be of type 'str'")
        raise TypeError

    if steps is None:
        if kmax is not None:
            try:
                kmax = float(kmax)
                assert(kmax>0)
            except Exception as e:
                log.exception("Argument 'kmax' needs to be a number > 0")
                raise
            if tmax is not None:
                log.exception("Arguments do not match: Please provide either"+\
                    " 'kmax' or 'tmin' and 'tmax' or 'steps'")
                raise TypeError
            else:
                tmax = kmax*dt
        if tmin is None:
            tmin = 1
        try:
            tmin=float(tmin)
            tmax=float(tmax)
            assert(tmin>=0 and tmax>tmin)
        except Exception as e:
            if kmax is None:
                log.exception("Arguments: 'tmax' and 'tmin' " +
                    "need to be floats with 'tmax' > 'tmin' >= 0")
            else:
                log.exception("Argument: 'kmax' is too small")

            raise
        steps = (int(tmin/dt), int(tmax/dt))
    else:
        if tmin is not None or tmax is not None or kmax is not None:
            log.exception("Arguments do not match: Please provide either "+\
                "'kmax' or 'tmin' and 'tmax' or 'steps'")
            raise TypeError
        log.debug("Argument 'steps' was provided to full_analysis()")

    defaultfits = False
    if fitfuncs is None:
        fitfuncs = ['e', 'eo']
        defaultfits = True
    elif isinstance(fitfuncs, str):
        fitfuncs = [fitfuncs]
    if not isinstance(fitfuncs, list):
        log.exception("Argument 'fitfuncs' needs to be of type 'str' or " +\
            "a list e.g. ['exponential', 'exponential_offset']")
        raise TypeError

    if targetplot is not None \
    and not isinstance(targetplot, matplotlib.axes.Axes):
        log.exception("Optional argument 'targetplot' needs " +
            "to be an instance of 'matplotlib.axes.Axes'")
        raise TypeError

    if title is not None:
        title = str(title)

    # as of v0.1.6 we decided to choose the default method based on the number of trials
    src = input_handler(data)

    if coefficientmethod is None and method is not None:
        coefficientmethod = method
    elif coefficientmethod is not None and method is not None:
        log.exception(
            "Keywords 'method' and 'coefficientmethod' are synonymous.\n" +
            "Provide one or the other!"
        )
        raise ValueError

    # for one trial the two methods are equal
    if coefficientmethod is None and src.shape[0] == 1:
        coefficientmethod = 'stationarymean'  # redundant with coefficients()
    elif coefficientmethod is None and src.shape[0] > 1:
        log.exception(
            "The provided data seems to have more than one trial. " +
            "Please specify a 'coefficientmethod':\n" +
            "'trialseparated' --- if your trials are long, or\n" +
            "'stationarymean' --- if you are sure that activity is stationary " +
            "across trials.\n" +
            "If you are unsure, we suggest to compare results from both methods."
            )
        raise TypeError

    if coefficientmethod not in [
    'trialseparated', 'ts', 'stationarymean', 'sm']:
        log.exception("Optional argument 'coefficientmethod' needs " +
            "to be either 'trialseparated' or 'stationarymean'")
        raise TypeError

    if (ut._log_locals):
        log.debug('Finished argument check. Locals: {}'.format(locals()))

    # ------------------------------------------------------------------ #
    # Continue with trusted arguments
    # ------------------------------------------------------------------ #

    if substracttrialaverage:
        src = src - np.mean(src, axis=0)

    log.debug('full_analysis() seeding to {}'.format(seed))
    if seed is None or seed == 'random':
        rkseed = seed
        ftseed = seed
    else:
        rkseed = seed*5330
        ftseed = seed*101

    if numboot == 'auto':
        nbt = 100
    else:
        nbt = numboot
    rks = coefficients(
        data=src,
        steps=steps,
        dt=dt,
        dtunit=dtunit,
        method=coefficientmethod,
        numboot=nbt,
        seed=rkseed,
    )

    fits = []
    for f in fitfuncs:
        if numboot == 'auto':
            if fitfunc_check(f) is f_exponential or \
                fitfunc_check(f) is f_exponential_offset:
                nbt = 100
            elif fitfunc_check(f) is f_complex:
                nbt = 0
            else:
                nbt = 100
        else:
            nbt = numboot
        fits.append(fit(data=rks, fitfunc=f, steps=steps,
            numboot=nbt, seed=ftseed))

    # ------------------------------------------------------------------ #
    # Output and Consistency Checks
    # ------------------------------------------------------------------ #

    warning = None
    if defaultfits:
        shownfits = [fits[1]]

        # no trials, no confidence
        if src.shape[0] == 1:
            warning = 'Not enough trials to calculate confidence intervals.'

        # check that tau  from exp and exp_off
        elif not ut._c_fits_consistent(fits[0], fits[1]):
            # warning = 'Exponential with offset resulted in ' + \
            #     '$\\tau = {:.2f}$ {}'.format(fits[1].tau, fits[1].dtunit)
            warning = 'Results from other fits differed beyond confidence.\n'+\
                "Try the 'fitfuncs' argument!"
    else:
        shownfits = fits
        warning = None

    # we do not want to have any python figures poping up when showoverivew
    # is set to false.
    # also, interactive: true seems to leak memory when many figures are opened
    with rc_context(rc={'interactive': showoverview}):
        if showoverview or saveoverview:
            panel = overview(src, [rks], shownfits, title=title,
                warning=warning, interactive=showoverview)

        res = OutputHandler([rks]+shownfits, ax=targetplot)

        if targetdir is not None:
            if (title is not None and title != ''):
                res.save(targetdir+"/"+title)
                if saveoverview:
                    panel.savefig(targetdir+"/"+title+"_overview.pdf")
            else:
                res.save(targetdir+"/full_analysis")
                if saveoverview:
                    panel.savefig(targetdir+"/full_analysis_overview.pdf")

        if showoverview:
            panel.show()
        else:
            # if interactive mode is on, panel would still be shown
            try:
                plt.close(panel)
            except:
                log.debug('Exception passed', exc_info=True)

        try:
            log.removeHandler(loghandler)
        except:
            log.debug('No handler to remove')

        log.info("full_analysis() done")
        return res



import os
import logging
import logging.handlers
import stat
import tempfile
import platform
import getpass
import math

import numpy as np
import scipy
import scipy.stats

__version__ = "unknown"
from ._version import __version__

log = logging.getLogger(__name__)
_targetdir = None
_logfilehandler = None
_logstreamhandler = None
_log_locals = False
_log_trace = True  # keep this on. seriously.
_print_progress = True
_keep_progress = False

# ------------------------------------------------------------------ #
# helper functions
# ------------------------------------------------------------------ #

def _set_permissions(fname, permissions=None):

    try:
        log.debug('Trying to set permissions of %s (%s) to %s',
            fname,
            oct(os.stat(fname)[stat.ST_MODE])[-3:],
            'defaults' if permissions is None else str(permissions))
        dirusr = os.path.abspath(os.path.expanduser('~'))
        if permissions is None:
            if not fname.startswith(dirusr):
                os.chmod(fname, 0o777)
        else:
            os.chmod(fname, int(str(permissions), 8))
    except Exception as e:
        log.debug('Unable set permissions of {}'.format(fname))

    log.debug('%s (now) has permissions %s',
        fname, oct(os.stat(fname)[stat.ST_MODE])[-3:])

def set_targetdir(fname=None, permissions=None):
    """
        set the global variable _targetdir.
        Per default, global log file is placed here.
        (and, in the future, any output when no path is specified)
        If no argument provided we try the default tmp directory.
        If permissions are not provided, uses 777 if not in user folder
        and defaults otherwise.
    """

    dirtmpsys = '/tmp' if platform.system() == 'Darwin' \
        else tempfile.gettempdir()

    if fname is None:
        fname = '{}/mre_{}'.format(dirtmpsys, getpass.getuser())
    else:
        try:
            fname = os.path.abspath(os.path.expanduser(fname))
        except Exception as e:
            log.debug('Specified file name caused an exception, using default',
                exc_info=True)
            fname = '{}/mre_{}'.format(dirtmpsys, getpass.getuser())

    log.debug('Setting global target directory to %s', fname)

    fname += '/'
    os.makedirs(fname, exist_ok=True)

    _set_permissions(fname, permissions)

    global _targetdir
    _targetdir = fname

    log.debug('Target directory set to %s', _targetdir)

def _c_rk_greater_zero(data, plim=0.1):
    """
        check rk are signigicantly larger than 0

        returns True if the test passed (and the null hypothesis was rejected)
    """
    if not isinstance(data, CoefficientResult):
        log.exception('_c_rk_greater_zero needs a CoefficientResult')
        raise TypeError

    # two sided
    t, p = scipy.stats.ttest_1samp(data.coefficients, 0.0)
    passed = False

    if p/2 < plim and t > 0:
        passed = True
    return passed, t, p

def _c_rk_smaller_one(data, plim=0.1):
    """
        check rk are signigicantly smaller than 1, this should fail if m>1

        returns True if the test passed (and the null hypothesis was rejected)
    """
    if not isinstance(data, CoefficientResult):
        log.exception('_c_rk_smaller_one needs a CoefficientResult')
        raise TypeError

    # two sided
    t, p = scipy.stats.ttest_1samp(data.coefficients, 1.0)
    passed = False

    if p/2 < plim and t < 0:
        passed = True
    return passed, t, p

def _c_fits_consistent(fit1, fit2, quantile=.125):
    """
        Check if fit1 and fit2 agree within confidence intervals.
        default quantile=.125 is between 1 and 2 sigma
    """
    # [.125, .25, .4, .5, .6, .75, .875]
    try:
        log.debug('Checking fits against each others {} quantiles' \
            .format(quantile))

        qmin = list(fit1.quantiles).index(  quantile)
        qmax = list(fit1.quantiles).index(1-quantile)
        log.debug('{} < {} < {} ?'.format(
            fit2.tauquantiles[qmin], fit1.tau, fit2.tauquantiles[qmax]))
        if fit1.tau > fit2.tauquantiles[qmax] \
        or fit1.tau < fit2.tauquantiles[qmin]:
            return False

        qmin = list(fit2.quantiles).index(  quantile)
        qmax = list(fit2.quantiles).index(1-quantile)
        log.debug('{} < {} < {} ?'.format(
            fit1.tauquantiles[qmin], fit2.tau, fit1.tauquantiles[qmax]))
        if fit2.tau > fit1.tauquantiles[qmax] \
        or fit2.tau < fit1.tauquantiles[qmin]:
            return False

        return True
    except Exception as e:
        log.debug('Quantile not found in fit', exc_info=True)

        return False

def _intersecting_index(ar1, ar2):
    """
        find indices where ar1 and ar2 have same elements. assumes uniqueness
        of elements in ar1 and ar2. returns (indices in ar1, indices in ar2)
    """
    try:
        _, comm1, comm2 = np.intersect1d(ar1, ar2, return_indices=True)
        # return_indices option needs numpy 1.15.0
    except TypeError:
        comm1 = []
        comm2 = []
        for idx, i in enumerate(ar1):
            for jdx, j in enumerate(ar2):
                if i == j:
                    comm1.append(idx)
                    comm2.append(jdx)
                    break
        comm1 = np.sort(comm1)
        comm2 = np.sort(comm2)

    return comm1, comm2

def _at_index(data, indices, keepdim=None, padding=np.nan):
    """
        get data[indices] safely and with optional padding to either
        indices.size or data.size
    """
    if not (keepdim is None or keepdim in ['data', 'index']):
        raise TypeError('unexpected argument keepdim={}'.format(keepdim))

    data    = np.asarray(data)
    indices = np.asarray(indices)
    i       = indices[indices < data.size]

    if keepdim is None:
        return data[i]
    elif keepdim == 'data':
        res = np.full(data.size, padding)
        res[i] = data[i]
        return res
    elif keepdim == 'index':
        res = np.full(indices.size, padding)
        if i.size !=0:
            res[0:indices.size-1] = data[i]
        return res

def _printeger(f, maxprec=5):
    try:
        f = float(f)
    except TypeError:
        log.debug('Exception when casting float in _printerger', exc_info=True)
        return 'None'
    prec=0
    while(not float(f*10**(prec)).is_integer() and prec <maxprec):
        prec+=1
    return str('{:.{p}f}'.format(f, p=prec))

def _prerror(f, ferr, errprec=2, maxprec=5):
    try:
        f = float(f)
    except TypeError:
        log.debug('Exception when casting float in _prerror', exc_info=True)
        return 'None'
    if ferr is None or ferr == 0:
        return _printeger(f, maxprec)
    if ferr < 1:
        prec = math.ceil(-math.log10(math.fabs(ferr) -
            math.fabs(math.floor(ferr)))) - 1
        return str('{:.{p}f}({:.0f})'.format(f, ferr*10**(prec+errprec),
            p=prec+errprec))
    else:
        return str('{}({})'.format(
            _printeger(f, errprec), _printeger(ferr, errprec)))


def math_from_doc(fitfunc, maxlen=np.inf):
    """convert sphinx compatible math to matplotlib/tex"""
    try:
        res = fitfunc.__doc__
        res = res.replace(':math:', '')
        res = res.replace('`', '$')
        if len(res) > maxlen:
            term = res.find(" + ", 0, len(res))
            res = res[:term+2]+' ...$'

        if len(res) > maxlen:
            if fitfunc.__name__ == 'f_complex':
                res = 'Complex'
            elif fitfunc.__name__ == 'f_exponential_offset':
                res = 'Exp+Offset'
            elif fitfunc.__name__ == 'f_exponential':
                res = 'Exponential'
            elif fitfunc.__name__ == 'f_linear':
                res = 'Linear'
            else:
                res = fitfunc.__name__

    except Exception as e:
        log.debug('Exception passed when casting function description',
            exc_info=True)
        res = fitfunc.__name__

    return res


# ------------------------------------------------------------------ #
# logging
# ------------------------------------------------------------------ #

# import tqdm if available and set some defaults
try:
    from tqdm import tqdm as __tqdm

    # we want to overload for custom defaults
    def tqdm(*args, **kwargs):

        if 'bar_format' not in kwargs:
            kwargs = dict(kwargs,
                bar_format="PROGRESS {percentage:3.0f}%|{bar}" +
                    "| {n_fmt}/{total_fmt} [{elapsed}<{remaining}]",
            )

        if 'leave' not in kwargs:
            kwargs = dict(kwargs, leave=True)

        if not _keep_progress:
            kwargs = dict(kwargs, leave=False)

        # this gives us two ways to disable printing of the progress bar.
        # globally via ut._print_progress and for individual function calls via disable
        if not _print_progress:
            kwargs = dict(kwargs, disable=True)

        return __tqdm(*args, **kwargs)

except ImportError:
    def tqdm(*args, **kwargs):
        if args:
            return args[0]
        return kwargs.get('iterable', None)

class CustomExceptionFormatter(logging.Formatter, object):

    def __init__(self,
        *args,
        # this is needed since i have not found an easy way to disable the
        # printing of the stack trace to console.
        force_disable_trace=False,
        indent_after_newline=0,
        **kwargs):
            self._force_disable_trace  = force_disable_trace
            self._padding = ' '*indent_after_newline
            super(CustomExceptionFormatter, self).__init__(*args, **kwargs)


    def format(self, record):
        # we want to indent after a newline
        # seems that this also indents exceptions/traces
        s = super(CustomExceptionFormatter, self).format(record)
        s = s.replace('\n', '\n'+self._padding)
        return s

    def formatException(self, exc_info):
        # original formatted exception
        exc_text = super(CustomExceptionFormatter, self).formatException(exc_info)
        # if not self._log_locals:
        if not _log_locals:
            # avoid printing 'NoneType' calling log.exception wihout try
            if exc_info[0] is None \
            or not _log_trace \
            or self._force_disable_trace:
                exc_text = ''
            else:
                exc_text = '\t'+exc_text.replace('\n', '\n\t')
            # k = exc_text.rfind('\n')
            # if k != -1:
                # exc_text = exc_text[:k]
                # pass

            return exc_text
        res = []
        # outermost frame of the traceback
        tb = exc_info[2]
        try:
            while tb.tb_next:
                tb = tb.tb_next  # Zoom to the innermost frame.
            res.append('Locals (most recent call last):')

            for k, v in tb.tb_frame.f_locals.items():
                res.append('  \'{}\': {}'.format(k, v))

            if not self._force_disable_trace and _log_trace:
                res.append(exc_text)

            res = '\t'+'\n'.join(res).replace('\n', '\n\t')
            # k = res.rfind('\n')
            # if k != -1:
                # res = res[:k]

            return res
        except:
            return ''

def _enable_detailed_logging():
    _log_locals = True
    _log_trace = True
    _logfilehandler.setLevel('DEBUG')
    _logstreamhandler.setLevel('DEBUG')
    logging.getLogger('py.warnings').addHandler(_logstreamhandler)
    log.debug('Logging set to full details, logs are saved at {}'.format(
        _logfilehandler.baseFilename))

def _exception_test(kwargs):
    mykwargs = kwargs
    localvar = 'dummy'
    try:
        i = 1/0
    except Exception as e:
        log.exception('_cause_exception try')
        raise Exception from e

def set_logfile(fname, loglevel=None, permissions=None):
    """
        Set the path where the global file logger writes the output.
        If the file is not within the user folder, permissions are set to 777
    """
    try:
        fname = os.path.abspath(os.path.expanduser(fname))
        if loglevel is None:
            loglevel = logging.getLevelName(_logfilehandler.level)
        log.debug('Setting log file to %s and level %s', fname, str(loglevel))
        _logfilehandler.setLevel(logging.getLevelName(loglevel))
    except Exception as e:
        log.debug(
            'Could not set loglevel. Maybe _logfilehandler doesnt exist yet?',
            exc_info=True)

    tempdir = os.path.abspath(fname+"/../")
    os.makedirs(tempdir, exist_ok=True)
    # _set_permissions(tempdir, permissions)
    # not sure yet if we want to overwrite permissions on possibly existing
    # or system directories. user can call _set_permissions() manually

    try:
        _logfilehandler.close()
        _logfilehandler.baseFilename = fname
        _set_permissions(fname, permissions)

        log.debug('Log file set to %s with level %s',
            _logfilehandler.baseFilename, str(loglevel))
    except Exception as e:
        log.debug('Could not set logfile. Maybe _logfilehandler doesnt exist?',
            exc_info=True)

def initialize():

    try:
        log.setLevel(logging.DEBUG)

        # create (global) file handler which logs even debug messages
        global _logfilehandler
        set_targetdir()

        # _logfilehandler = logging.FileHandler(_targetdir+'mre.log', 'w')
        _logfilehandler = logging.handlers.RotatingFileHandler(_targetdir+'mre.log',
            mode='w', maxBytes=50*1024*1024, backupCount=9)
        _set_permissions(_targetdir+'mre.log')
        _logfilehandler.setFormatter(CustomExceptionFormatter(
            '%(asctime)s %(levelname)8s: %(message)s', "%Y-%m-%d %H:%M:%S",
             force_disable_trace=False,
            indent_after_newline=30))
        _logfilehandler.setLevel(logging.DEBUG)
        log.addHandler(_logfilehandler)

        # create (global) console handler with a higher log level
        global _logstreamhandler
        _logstreamhandler = logging.StreamHandler()
        _logstreamhandler.setFormatter(CustomExceptionFormatter(
            '%(levelname)-8s %(message)s',
            force_disable_trace=False,
            indent_after_newline=9))
        _logstreamhandler.setLevel(logging.INFO)
        log.addHandler(_logstreamhandler)

        log.info('Loaded mrestimator v%s, writing to %s',
            __version__, _targetdir)

        # capture (numpy) warnings and only log them to the global log file
        logging.captureWarnings(True)
        logging.getLogger('py.warnings').addHandler(_logfilehandler)
        # logging.getLogger('py.warnings').addHandler(_logstreamhandler)

    except Exception as e:
        print('Loaded mrestimator v{}, but logger could not be set up for {}'
            .format(__version__, _targetdir))
        print(e)



__version__ = "unknown"
from ._version import __version__


from mrestimator import utility as ut
ut.initialize()
log = ut.log

from .coefficients import CoefficientResult, coefficients
from .fit          import *
from .input_output import *
from .simulate     import *
from .wrapper      import *























import re
import os
import glob
import inspect
import logging

from mrestimator import utility as ut

log = ut.log
from mrestimator import CoefficientResult
from mrestimator import FitResult
from mrestimator import __version__

import numpy as np
import matplotlib

if os.environ.get("DISPLAY", "") == "":
    log.info(
        "No display found. Using non-interactive Agg backend for plotting. " +
        "Check your $DISPLAY environment variable."
    )
    matplotlib.use("Agg")
import matplotlib.pyplot as plt


def input_handler(items, **kwargs):
    """
        Helper function that attempts to detect provided input and convert it
        to the format used by the toolbox. Ideally, you provide the native
        format, a :class:`numpy.ndarray` of ``shape(numtrials, datalength)``.

        *Not implemented yet*:
        All trials should have the same data length, otherwise they will be
        padded.

        The toolbox uses two dimensional `ndarrays` for
        providing the data to/from functions. This allows to
        consistently access trials and data via the first and second index,
        respectively.

        Parameters
        ----------
        items : str, list or ~numpy.ndarray
            A `string` is assumed to be the path to
            file that is then imported as pickle or plain text.
            Wildcards should work.
            Alternatively, you can provide a `list` or `ndarray` containing
            strings or already imported data. In the latter case,
            `input_handler` attempts to convert it to the right format.

        kwargs
            Keyword arguments passed to :func:`numpy.loadtxt` when filenames
            are detected (see numpy documentation for a full list).
            For instance, you can provide ``usecols=(1,2)``
            if your files have multiple columns and only the column 1 and 2
            contain trial data you want to use.
            The input handler adds each column in each file to the list of
            trials.

        Returns
        -------
        : :class:`~numpy.ndarray`
            containing your data (hopefully)
            formatted correctly. Access via ``[trial, datapoint]``

        Example
        -------
        .. code-block:: python

            # import a single file
            prepared = mre.input_handler('/path/to/yourfiles/trial_1.csv')
            print(prepared.shape)

            # or from a list of files
            myfiles = ['~/data/file_0.csv', '~/data/file_1.csv']
            prepared = mre.input_handler(myfiles)

            # all files matching the wildcard, but only columns 3 and 4
            prepared = mre.input_handler('~/data/file_*.csv', usecols=(3, 4))

            # access your data, e.g. measurement 10 of trial 3
            pt = prepared[3, 10]
        ..
    """
    invstr = (
        "\nInvalid input, please provide one of the following:\n"
        "\t- path to pickle or plain file as string,\n"
        '\t  wildcards should work "/path/to/filepattern*"\n'
        "\t- numpy array or list containing spike data or filenames\n"
    )

    log.debug("input_handler()")
    situation = -1
    # cast tuple to list, maybe this can be done for other types in the future
    if isinstance(items, tuple):
        log.debug("input_handler() detected tuple, casting to list")
        items = list(items)
    if isinstance(items, np.ndarray):
        if items.dtype.kind in ["i", "f", "u"]:
            log.info("input_handler() detected ndarray of numbers")
            situation = 0
        elif items.dtype.kind in ["S", "U"]:
            log.info("input_handler() detected ndarray of strings")
            situation = 1
            temp = set()
            for item in items.astype("U"):
                temp.update(glob.glob(os.path.expanduser(item)))
            if len(items) != len(temp):
                log.debug(
                    "{} duplicate files were excluded".format(len(items) - len(temp))
                )
            items = temp
        else:
            log.exception("Numpy.ndarray is neither data nor file path.%s", invstr)
            raise ValueError
    elif isinstance(items, list):
        if all(isinstance(item, str) for item in items):
            log.info("input_handler() detected list of strings")
            try:
                log.debug("Parsing to numpy ndarray as float")
                items = np.asarray(items, dtype=float)
                situation = 0
            except Exception as e:
                log.debug("Exception caught, parsing as file path", exc_info=True)
            situation = 1
            temp = set()
            for item in items:
                temp.update(glob.glob(os.path.expanduser(item)))
            if len(items) != len(temp):
                log.debug(
                    "{} duplicate files were excluded".format(len(items) - len(temp))
                )
            items = temp
        elif all(isinstance(item, np.ndarray) for item in items):
            log.info("input_handler() detected list of ndarrays")
            situation = 0
        else:
            try:
                log.info(
                    "input_handler() detected list, "
                    + "parsing to numpy ndarray as float"
                )
                situation = 0
                items = np.asarray(items, dtype=float)
            except Exception as e:
                log.exception("%s", invstr)
                raise
    elif isinstance(items, str):
        log.info("input_handler() detected filepath '{}'".format(items))
        items = glob.glob(os.path.expanduser(items))
        situation = 1
    else:
        log.exception("Unknown argument type,%s", invstr)
        raise TypeError

    if situation == 0:
        retdata = np.stack((items), axis=0)
        if len(retdata.shape) == 1:
            retdata = retdata.reshape((1, len(retdata)))
    elif situation == 1:
        if len(items) == 0:
            # glob of earlyier analysis returns nothing if file not found
            log.exception(
                "Specifying absolute file path is recommended, "
                + "input_handler() was looking in {}\n".format(os.getcwd())
                + "\tUse 'os.chdir(os.path.dirname(__file__))' to set the "
                + "working directory to the location of your script file"
            )
            raise FileNotFoundError

        data = []
        for idx, item in enumerate(items):
            try:
                log.debug("Loading with np.loadtxt: {}".format(item))
                if "unpack" in kwargs and not kwargs.get("unpack"):
                    log.warning(
                        "Argument 'unpack=False' is not recommended,"
                        + " data is usually stored in columns"
                    )
                else:
                    kwargs = dict(kwargs, unpack=True)
                if "ndmin" in kwargs and kwargs.get("ndmin") != 2:
                    log.exception("Argument ndmin other than 2 not supported")
                    raise ValueError
                else:
                    kwargs = dict(kwargs, ndmin=2)
                # fix for numpy 1.11
                if "usecols" in kwargs and isinstance(kwargs.get("usecols"), int):
                    kwargs = dict(kwargs, usecols=[kwargs.get("usecols")])
                result = np.loadtxt(item, **kwargs)
                data.append(result)
            except Exception as e:
                log.debug(
                    "Exception caught, Loading with np.load " + "{}".format(item),
                    exc_info=True,
                )
                result = np.load(item)
                data.append(result)

        try:
            retdata = np.vstack(data)
        except ValueError:
            minlenx = min(l.shape[0] for l in data)
            minleny = min(l.shape[1] for l in data)

            log.debug(
                "Files have different length, resizing to shortest "
                "one ({}, {})".format(minlenx, minleny),
                exc_info=True,
            )
            for d, dat in enumerate(data):
                data[d] = np.resize(dat, (minlenx, minleny))
            retdata = np.vstack(data)

    else:
        log.exception("Unknown situation%s", invstr)
        raise NotImplementedError

    # final check
    if len(retdata.shape) == 2:
        log.info(
            "input_handler() returning ndarray with %d trial(s) and %d " + "datapoints",
            retdata.shape[0],
            retdata.shape[1],
        )
        return retdata
    else:
        log.warning(
            "input_handler() guessed data type incorrectly to shape "
            + "{}, please try something else".format(retdata.shape)
        )
        return retdata


class OutputHandler:
    """
        The OutputHandler can be used to export results and to
        create charts with
        timeseries, correlation-coefficients or fits.

        The main concept is to have one handler per plot. It contains
        functions to add content into an existing matplotlib axis (subplot),
        or, if not provided, creates a new figure.
        Most importantly, it also exports plaintext of the respective source
        material so figures are reproducible.

        Note: If you want to have a live preview of the figures that are
        automatically generated with matplotlib, you HAVE to assign the result
        of `mre.OutputHandler()` to a variable. Otherwise, the created figures
        are not retained and vanish instantly.

        Attributes
        ----------
        rks: list
            List of the :obj:`CoefficientResult`. Added with `add_coefficients()`

        fits: list
            List of the :obj:`FitResult`. Added with `add_fit()`

        Example
        -------
        .. code-block:: python

            import numpy as np
            import matplotlib.pyplot as plt
            import mrestimator as mre

            bp  = mre.simulate_branching(numtrials=15)
            rk1 = mre.coefficients(bp, method='trialseparated',
                desc='T')
            rk2 = mre.coefficients(bp, method='stationarymean',
                desc='S')

            m1 = mre.fit(rk1)
            m2 = mre.fit(rk2)

            # create a new handler by passing with list of elements
            out = mre.OutputHandler([rk1, m1])

            # manually add elements
            out.add_coefficients(rk2)
            out.add_fit(m2)

            # save the plot and meta to disk
            out.save('~/test')
        ..

        Working with existing figures:

        .. code-block:: python

            # create figure with subplots
            fig = plt.figure()
            ax1 = fig.add_subplot(221)
            ax2 = fig.add_subplot(222)
            ax3 = fig.add_subplot(223)
            ax4 = fig.add_subplot(224)

            # show each chart in its own subplot
            mre.OutputHandler(rk1, ax1)
            mre.OutputHandler(rk2, ax2)
            mre.OutputHandler(m1, ax3)
            mre.OutputHandler(m2, ax4)

            # matplotlib customisations
            myaxes = [ax1, ax2, ax3, ax4]
            for ax in myaxes:
                ax.spines['top'].set_visible(False)
                ax.spines['right'].set_visible(False)

            plt.show(block=False)

            # hide a legend
            ax1.legend().set_visible(False)
            plt.draw()
        ..
    """

    def __init__(self, data=None, ax=None):
        """
            Construct a new OutputHandler, optionally you can provide
            the a list of elements to plot.

            ToDo: Make the OutputHandler talk to each other so that
            when one is written (possibly linked to others via one figure)
            all subfigure meta data is exported, too.

            Parameters
            ----------
            data : list, CoefficientResult or FitResult, optional
                List of the elements to plot/export. Can be added later.

            ax : ~matplotlib.axes.Axes, optional
                The an instance of a matplotlib axes (a subplot) to plot into.
        """
        if isinstance(ax, matplotlib.axes.Axes):
            self.ax = ax
            self.axshared = True
        elif ax is None:
            self.axshared = False
            # fig = plt.figure()
            # self.ax = fig.add_subplot(111, rasterized=True)
            _, self.ax = plt.subplots()
            # everything below zorder 0 gets rastered to one layer
            self.ax.set_rasterization_zorder(0)
        else:
            log.exception(
                "Argument 'ax' provided to OutputHandler is not "
                + " an instance of matplotlib.axes.Axes\n"
                + "\tIn case you want to add multiple items, pass them in a list "
                + "as the first argument"
            )
            raise TypeError

        self.rks = []
        self.rklabels = []
        self.rkcurves = []
        self.rkkwargs = []
        self.fits = []
        self.fitlabels = []
        self.fitcurves = []  # list of lists of drawn curves for each fit
        self.fitkwargs = []
        self.dt = 1
        self.dtunit = None
        self.type = None
        self.xdata = None
        self.ydata = []  # list of 1d np arrays
        self.xlabel = None
        self.ylabels = []

        # single argument to list
        if (
            isinstance(data, CoefficientResult)
            or isinstance(data, FitResult)
            or isinstance(data, np.ndarray)
        ):
            data = [data]

        for d in data or []:
            if isinstance(d, CoefficientResult):
                self.add_coefficients(d)
            elif isinstance(d, FitResult):
                self.add_fit(d)
            elif isinstance(d, np.ndarray):
                self.add_ts(d)
            else:
                log.exception(
                    "Please provide a list containing "
                    "\tCoefficientResults and/or FitResults\n"
                )
                raise ValueError

    def __del__(self):
        """
            close opened figures when outputhandler is no longer used
        """
        if not self.axshared:
            try:
                plt.close(self.ax.figure)
                # pass
            except Exception as e:
                log.debug("Exception passed", exc_info=True)

    def set_xdata(self, data=None, dt=1, dtunit=None):
        """
            Adjust xdata of the plot, matching the input value.
            Returns an array of indices matching the incoming indices to
            already present ones. Automatically called when adding content.

            If you want to customize the plot range, add all the content
            and use matplotlibs
            :obj:`~matplotlib.axes.Axes.set_xlim` function once at the end.
            (`set_xdata()` also manages meta data and can only *increase* the
            plot range)

            Parameters
            ----------
            data : ~numpy.array
                x-values to plot the fits for. `data` does not need to be
                spaced equally but is assumed to be sorted.

            dt : float
                check if existing data can be mapped to the new, provided `dt`
                or the other way around. `set_xdata()` pads
                undefined areas with `nan`.

            dtunit : str
                check if the new `dtunit` matches the one set previously. Any
                padding to match `dt` is only done if `dtunits` are the same,
                otherwise the plot falls back to using generic integer steps.

            Returns
            -------
            : :class:`~numpy.array`
                containing the indices where the `data` given to this function
                coincides with (possibly) already existing data that was
                added/plotted before.

            Example
            -------
            .. code-block:: python

                out = mre.OutputHandler()

                # 100 intervals of 2ms
                out.set_xdata(np.arange(0,100), dt=2, dtunit='ms')

                # increase resolution to 1ms for the first 50ms
                # this changes the existing structure in the meta data. also
                # the axis of `out` is not equally spaced anymore
                fiftyms = np.arange(0,50)
                out.set_xdata(fiftyms, dt=1, dtunit='ms')

                # data with larger intervals is less dense, the returned list
                # tells you which index in `out` belongs to every index
                # in `xdat`
                xdat = np.arange(0,50)
                ydat = np.random_sample(50)
                inds = out.set_xdata(xdat, dt=4, dtunit='ms')

                # to pad `ydat` to match the axis of `out`:
                temp = np.full(out.xdata.size, np.nan)
                temp[inds] = ydat

            ..
        """
        log.debug("OutputHandler.set_xdata()")
        # make sure data is not altered
        xdata = np.copy(data.astype("float64"))
        # xdata = data

        # nothing set so far, no arugment provided, return some default
        if self.xdata is None and xdata is None:
            self.xdata = np.arange(0, 1501)
            self.dtunit = dtunit
            self.dt = dt
            return np.arange(0, 1501)

        # set x for the first time, copying input
        if self.xdata is None:
            self.xdata = np.array(xdata)
            self.dtunit = dtunit
            self.dt = dt
            return np.arange(0, self.xdata.size)

        # no new data provided, no need to call this
        elif xdata is None:
            log.debug(
                "set_xdata() called without argument when "
                + "xdata is already set. Nothing to adjust"
            )
            return np.arange(0, self.xdata.size)

        # compare dtunits
        elif dtunit != self.dtunit and dtunit is not None:
            log.warning(
                "'dtunit' does not match across added elements, "
                + "adjusting axis label to '[different units]'"
            )
            regex = r"\[.*?\]"
            oldlabel = self.ax.get_xlabel()
            self.ax.set_xlabel(re.sub(regex, "[different units]", oldlabel))

        # set dtunit to new value if not assigned yet
        elif self.dtunit is None and dtunit is not None:
            self.dtunit = dtunit

        # new data matches old data, nothing to adjust
        if np.array_equal(self.xdata, xdata) and self.dt == dt:
            return np.arange(0, self.xdata.size)

        # compare timescales dt
        elif self.dt < dt:
            log.debug("dt does not match,")
            scd = dt / self.dt
            if float(scd).is_integer():
                log.debug(
                    "Changing axis values of new data (dt={})".format(dt)
                    + "to match higher resolution of "
                    + "old xaxis (dt={})".format(self.dt)
                )
                scd = dt / self.dt
                xdata *= scd
            else:
                log.warning(
                    "New 'dt={}' is not an integer multiple of ".format(dt)
                    + "the previous 'dt={}\n".format(self.dt)
                    + "Plotting with '[different units]'\n"
                    + "As a workaround, try adding the data with the "
                    + "smallest 'dt' first"
                )
                try:
                    regex = r"\[.*?\]"
                    oldlabel = self.ax.get_xlabel()
                    self.ax.set_xlabel(re.sub(regex, "[different units]", oldlabel))
                    self.xlabel = re.sub(regex, "[different units]", self.xlabel)
                except TypeError:
                    log.debug("Exception passed", exc_info=True)

        elif self.dt > dt:
            scd = self.dt / dt
            if float(scd).is_integer():
                log.debug(
                    "Changing 'dt' to new value 'dt={}'\n".format(dt)
                    + "\tAdjusting existing axis values (dt={})".format(self.dt)
                )
                self.xdata *= scd
                self.dt = dt
                try:
                    regex = r"\[.*?\]"
                    oldlabel = self.ax.get_xlabel()
                    if self.dt == 1:
                        newlabel = str("[{}]".format(self.dtunit))
                    else:
                        newlabel = str(
                            "[{} {}]".format(ut._printeger(self.dt), self.dtunit)
                        )
                    self.ax.set_xlabel(re.sub(regex, newlabel, oldlabel))
                    self.xlabel = re.sub(regex, newlabel, self.xlabel)
                except TypeError:
                    pass
            else:
                log.warning(
                    "old 'dt={}' is not an integer multiple ".format(self.dt)
                    + "of the new value 'dt={}'\n".format(self.dt)
                    + "\tPlotting with '[different units]'\n"
                )
                try:
                    regex = r"\[.*?\]"
                    oldlabel = self.ax.get_xlabel()
                    self.ax.set_xlabel(re.sub(regex, "[different units]", oldlabel))
                    self.xlabel = re.sub(regex, "[different units]", self.xlabel)
                except TypeError:
                    pass

        # check if new is subset of old
        temp = np.union1d(self.xdata, xdata)
        if not np.array_equal(self.xdata, temp):
            log.debug("Rearranging present data")
            _, indtemp = ut._intersecting_index(self.xdata, temp)
            self.xdata = temp
            for ydx, col in enumerate(self.ydata):
                coln = np.full(self.xdata.size, np.nan)
                coln[indtemp] = col
                self.ydata[ydx] = coln

        # return list of indices where to place new ydata in the existing
        # (higher-resolution) notation
        indold, indnew = ut._intersecting_index(self.xdata, xdata)
        assert len(indold) == len(xdata)

        return indold

    def add_coefficients(self, data, **kwargs):
        """
            Add an individual CoefficientResult. Note that it is not possible
            to add the same data twice, instead it will be redrawn with
            the new arguments/style options provided.

            Parameters
            ----------
            data : CoefficientResult
                Added to the list of plotted elements.

            kwargs
                Keyword arguments passed to
                :obj:`matplotlib.axes.Axes.plot`. Use to customise the
                plots. If a `label` is set via `kwargs`, it will be used to
                overwrite the description of `data` in the meta file.
                If an alpha value is or linestyle is set, the shaded error
                region will be omitted.

            Example
            -------
            .. code-block:: python

                rk = mre.coefficients(mre.simulate_branching())

                mout = mre.OutputHandler()
                mout.add_coefficients(rk, color='C1', label='test')
            ..
        """
        if not isinstance(data, CoefficientResult):
            log.exception("'data' needs to be of type CoefficientResult")
            raise ValueError
        if not (self.type is None or self.type == "correlation"):
            log.exception(
                "It is not possible to 'add_coefficients()' to "
                + "an OutputHandler containing a time series\n"
                + "\tHave you previously called 'add_ts()' on this handler?"
            )
            raise ValueError
        self.type = "correlation"

        # description for columns of meta data
        desc = str(data.desc)

        # plot legend label
        if "label" in kwargs:
            label = kwargs.get("label")
            if label == "":
                label = None
            if label is None:
                labelerr = None
            else:
                # user wants custom label not intended to hide the legend
                label = str(label)
                labelerr = str(label) + " Errors"
                # apply to meta data, too
                desc = str(label)
        else:
            # user has not set anything, copy from desc if set
            label = "Data"
            labelerr = "Errors"
            if desc != "":
                label = desc
                labelerr = desc + " Errors"

        if desc != "":
            desc += " "

        # dont put errors in the legend. this should become a user choice
        labelerr = ""

        # no previous coefficients present
        if len(self.rks) == 0:
            self.dt = data.dt
            self.dtunit = data.dtunit
            if self.dt == 1:
                self.xlabel = "steps[{}]".format(data.dtunit)
                self.ax.set_xlabel("k [{}]".format(data.dtunit))
            else:
                self.xlabel = "steps[{} {}]".format(
                    ut._printeger(data.dt, 5), data.dtunit
                )
                self.ax.set_xlabel(
                    "k [{} {}]".format(ut._printeger(data.dt, 5), data.dtunit)
                )
            self.ax.set_ylabel("$r_{k}$")
            self.ax.set_title("Correlation", fontweight="bold")

        # we dont support adding duplicates
        oldcurves = []
        if data in self.rks:
            indrk = self.rks.index(data)
            log.warning(
                "Coefficients ({}/{}) ".format(self.rklabels[indrk][0], label)
                + "have already been added\n\tOverwriting with new style"
            )
            del self.rks[indrk]
            del self.rklabels[indrk]
            oldcurves = self.rkcurves[indrk]
            del self.rkcurves[indrk]
            del self.rkkwargs[indrk]

        # add to meta data
        else:
            inds = self.set_xdata(data.steps, dt=data.dt, dtunit=data.dtunit)
            ydata = np.full(self.xdata.size, np.nan)
            ydata[inds] = data.coefficients
            self.ydata.append(ydata)
            self.ylabels.append(desc + "coefficients")

            if data.stderrs is not None:
                ydata = np.full(self.xdata.size, np.nan)
                ydata[inds] = data.stderrs
                self.ydata.append(ydata)
                self.ylabels.append(desc + "stderrs")

        self.rks.append(data)
        self.rklabels.append([label, labelerr])
        self.rkcurves.append(oldcurves)
        self.rkkwargs.append(kwargs)

        # refresh coefficients
        for r in self.rks:
            self._render_coefficients(r)

        # refresh fits
        for f in self.fits:
            self._render_fit(f)

    # need to implement using kwargs
    def _render_coefficients(self, rk):
        # (re)draw over (possibly) new xrange/dt
        indrk = self.rks.index(rk)
        label, labelerr = self.rklabels[indrk]
        kwargs = self.rkkwargs[indrk].copy()

        # reset curves and recover color
        color = None
        for idx, curve in enumerate(self.rkcurves[indrk]):
            if idx == 0:
                color = curve.get_color()
            curve.remove()
        self.rkcurves[indrk] = []

        if "color" not in kwargs:
            kwargs = dict(kwargs, color=color)
        if "zorder" not in kwargs:
            kwargs = dict(kwargs, zorder=1 + 0.01 * indrk)

        kwargs = dict(kwargs, label=label)

        # redraw plot
        (p,) = self.ax.plot(rk.steps * rk.dt / self.dt, rk.coefficients, **kwargs)
        self.rkcurves[indrk].append(p)

        try:
            if rk.stderrs is not None and "alpha" not in kwargs:
                err1 = rk.coefficients - rk.stderrs
                err2 = rk.coefficients + rk.stderrs
                kwargs.pop("color")
                kwargs.pop("zorder")
                kwargs = dict(
                    kwargs,
                    label=labelerr,
                    alpha=0.2,
                    facecolor=p.get_color(),
                    zorder=p.get_zorder() - 1,
                )
                d = self.ax.fill_between(
                    rk.steps * rk.dt / self.dt, err1, err2, **kwargs
                )
                self.rkcurves[indrk].append(d)
        # not all kwargs are compaible with fill_between
        except AttributeError:
            pass

        if label is not None:
            self.ax.legend()

        # confirm ticks, it's confusing that we should have a tick at k=0
        old_limit = self.ax.get_xlim()
        old_ticks = list(self.ax.get_xticks())
        new_ticks = [1] + [i for i in old_ticks if i > 1]
        self.ax.set_xticks(new_ticks)
        self.ax.set_xlim(old_limit)  # matplotlib might change xlim to match ticks

    def add_fit(self, data, **kwargs):
        """
            Add an individual FitResult. By default, the part of the fit that
            contributed to the fitting is drawn solid, the remaining range
            is dashed. Note that it is not possible
            to add the same data twice, instead it will be redrawn with
            the new arguments/style options provided.

            Parameters
            ----------
            data : FitResult
                Added to the list of plotted elements.

            kwargs
                Keyword arguments passed to
                :obj:`matplotlib.axes.Axes.plot`. Use to customise the
                plots. If a `label` is set via `kwargs`, it will be added
                as a note in the meta data. If `linestyle` is set, the
                dashed plot of the region not contributing to the fit is
                omitted.
        """
        if not isinstance(data, FitResult):
            log.exception("'data' needs to be of type FitResult")
            raise ValueError
        if not (self.type is None or self.type == "correlation"):
            log.exception(
                "It is not possible to 'add_fit()' to "
                + "an OutputHandler containing a time series\n"
                + "\tHave you previously called 'add_ts()' on this handler?"
            )
            raise ValueError
        self.type = "correlation"

        if self.xdata is None:
            self.dt = data.dt
            self.dtunit = data.dtunit
            self.ax.set_xlabel("k [{} {}]".format(data.dt, data.dtunit))
            self.ax.set_ylabel("$r_{k}$")
            self.ax.set_title("Correlation", fontweigh="bold")
        inds = self.set_xdata(data.steps, dt=data.dt, dtunit=data.dtunit)

        # description for fallback
        desc = str(data.desc)

        # plot legend label
        if "label" in kwargs:
            label = kwargs.get("label")
            if label == "":
                label = None
            else:
                # user wants custom label not intended to hide the legend
                label = str(label)
        else:
            # user has not set anything, copy from desc if set
            label = "Fit " + ut.math_from_doc(data.fitfunc, 0)
            if desc != "":
                label = desc + " " + label

        # we dont support adding duplicates
        oldcurves = []
        if data in self.fits:
            indfit = self.fits.index(data)
            log.warning(
                "Fit was already added ({})\n".format(self.fitlabels[indfit])
                + "\tOverwriting with new style"
            )
            del self.fits[indfit]
            del self.fitlabels[indfit]
            oldcurves = self.fitcurves[indfit]
            del self.fitcurves[indfit]
            del self.fitkwargs[indfit]

        self.fits.append(data)
        self.fitlabels.append(label)
        self.fitcurves.append(oldcurves)
        self.fitkwargs.append(kwargs)

        # refresh coefficients
        for r in self.rks:
            self._render_coefficients(r)

        # refresh fits
        for f in self.fits:
            self._render_fit(f)

    def _render_fit(self, fit):
        # (re)draw fit over (possibly) new xrange
        indfit = self.fits.index(fit)
        label = self.fitlabels[indfit]
        kwargs = self.fitkwargs[indfit].copy()
        color = None
        for idx, curve in enumerate(self.fitcurves[indfit]):
            if idx == 0:
                color = curve.get_color()
            curve.remove()
        self.fitcurves[indfit] = []

        if "color" not in kwargs:
            kwargs = dict(kwargs, color=color)
        if "zorder" not in kwargs:
            kwargs = dict(kwargs, zorder=4 + 0.01 * indfit)

        kwargs = dict(kwargs, label=label)

        # update plot
        (p,) = self.ax.plot(
            fit.steps * fit.dt / self.dt,
            fit.fitfunc(fit.steps * fit.dt, *fit.popt),
            **kwargs,
        )
        self.fitcurves[indfit].append(p)

        # only draw dashed not-fitted range if no linestyle is specified
        if fit.steps[0] > self.xdata[0] or fit.steps[-1] < self.xdata[-1]:
            if "linestyle" not in kwargs and "ls" not in kwargs:
                kwargs.pop("label")
                kwargs = dict(kwargs, ls="dashed", color=p.get_color())
                (d,) = self.ax.plot(
                    self.xdata, fit.fitfunc(self.xdata * self.dt, *fit.popt), **kwargs
                )
                self.fitcurves[indfit].append(d)

        # errors as shaded area
        if False:
            try:
                if fit.taustderr is not None and "alpha" not in kwargs:
                    ptmp = np.copy(fit.popt)
                    ptmp[0] = fit.tau - fit.taustderr
                    err1 = fit.fitfunc(self.xdata * self.dt, *ptmp)
                    ptmp[0] = fit.tau + fit.taustderr
                    err2 = fit.fitfunc(self.xdata * self.dt, *ptmp)
                    kwargs.pop("color")
                    kwargs.pop("label")
                    kwargs = dict(
                        kwargs,
                        alpha=0.2,
                        facecolor=p.get_color(),
                        zorder=0 + 0.01 * indfit,
                    )
                    s = self.ax.fill_between(self.xdata, err1, err2, **kwargs)
                    self.fitcurves[indfit].append(s)
            # not all kwargs are compaible with fill_between
            except AttributeError:
                log.debug("Exception passed", exc_info=True)

        if label is not None:
            self.ax.legend()

        # confirm ticks, it's confusing that we should have a tick at k=0
        old_limit = self.ax.get_xlim()
        old_ticks = list(self.ax.get_xticks())
        new_ticks = [1] + [i for i in old_ticks if i > 1]
        self.ax.set_xticks(new_ticks)
        self.ax.set_xlim(old_limit)  # matplotlib might change xlim to match ticks

    def add_ts(self, data, **kwargs):
        """
            Add timeseries (possibly with trial structure).
            Not compatible with OutputHandlers that have data added via
            `add_fit()` or `add_coefficients()`.

            Parameters
            ----------
            data : ~numpy.ndarray
                The timeseries to plot. If the `ndarray` is two dimensional,
                a trial structure is assumed and all trials are plotted using
                the same style (default or defined via `kwargs`).
                *Not implemented yet*: Providing a ts with its own custom axis

            kwargs
                Keyword arguments passed to
                :obj:`matplotlib.axes.Axes.plot`. Use to customise the
                plots.

            Example
            -------
            .. code-block:: python

                bp = mre.simulate_branching(numtrials=10)

                tsout = mre.OutputHandler()
                tsout.add_ts(bp, alpha=0.1, label='Trials')
                tsout.add_ts(np.mean(bp, axis=0), label='Mean')

                plt.show()
            ..
        """
        if not (self.type is None or self.type == "timeseries"):
            log.exception(
                "Adding time series 'add_ts()' is not "
                + "compatible with an OutputHandler that has coefficients\n"
                + "\tHave you previously called 'add_coefficients()' or "
                + "'add_fit()' on this handler?"
            )
            raise ValueError
        self.type = "timeseries"
        if not isinstance(data, np.ndarray):
            data = np.array(data)
        if len(data.shape) < 2:
            data = data.reshape((1, len(data)))
        elif len(data.shape) > 2:
            log.exception("Only compatible with up to two dimensions")
            raise NotImplementedError

        desc = kwargs.get("label") if "label" in kwargs else "ts"
        color = kwargs.get("color") if "color" in kwargs else None
        alpha = kwargs.get("alpha") if "alpha" in kwargs else None
        # per default, if more than one series provided reduce alpha
        if data.shape[0] > 1 and not "alpha" in kwargs:
            alpha = 0.1
        kwargs = dict(kwargs, alpha=alpha)

        if "zorder" not in kwargs:
            kwargs = dict(kwargs, zorder=-1)

        for idx, dat in enumerate(data):
            if self.xdata is None:
                self.set_xdata(np.arange(1, data.shape[1] + 1))
                self.xlabel = "timesteps"
                self.ax.set_xlabel("t")
                self.ax.set_ylabel("$A_{t}$")
                self.ax.set_title("Time Series", fontweight="bold")
            elif len(self.xdata) != len(dat):
                log.exception("Time series have different length")
                raise NotImplementedError
            # if self.ydata is None:
            #     self.ydata = np.full((1, len(self.xdata)), np.nan)
            #     self.ydata[0] = dat
            # else:
            #     self.ydata = np.vstack((self.ydata, dat))
            self.ydata.append(dat)

            self.ylabels.append(desc + "[{}]".format(idx) if len(data) > 1 else desc)
            (p,) = self.ax.plot(self.xdata, dat, **kwargs)

            # dont plot an empty legend
            if kwargs.get("label") is not None and kwargs.get("label") != "":
                self.ax.legend()

            # only add to legend once
            if idx == 0:
                kwargs = dict(kwargs, label=None)
                kwargs = dict(kwargs, color=p.get_color())

    def save(self, fname="", ftype="pdf", dpi=300):
        """
            Saves plots (ax element of this handler) and source that it was
            created from to the specified location.

            Parameters
            ----------
            fname : str, optional
                Path where to save, without file extension. Defaults to "./mre"
        """
        self.save_plot(fname, ftype=ftype, dpi=dpi)
        self.save_meta(fname)

    def save_plot(self, fname="", ftype="pdf", dpi=300):
        """
            Only saves plots (ignoring the source) to the specified location.

            Parameters
            ----------
            fname : str, optional
                Path where to save, without file extension. Defaults to "./mre"

            ftype: str, optional
                So far, only 'pdf' and 'png' are implemented.
        """
        if not isinstance(fname, str):
            fname = str(fname)
        if fname == "":
            fname = "./mre"

        # try creating enclosing dir if not existing
        tempdir = os.path.abspath(os.path.expanduser(fname + "/../"))
        os.makedirs(tempdir, exist_ok=True)

        fname = os.path.expanduser(fname)

        if isinstance(ftype, str):
            ftype = [ftype]
        for t in list(ftype):
            log.info("Saving plot to {}.{}".format(fname, t.lower()))
            if t.lower() == "pdf":
                self.ax.figure.savefig(fname + ".pdf", dpi=dpi)
            elif t.lower() == "png":
                self.ax.figure.savefig(fname + ".png", dpi=dpi)
            else:
                log.exception("Unsupported file format '{}'".format(t))
                raise ValueError

    def save_meta(self, fname=""):
        """
            Saves only the details/source used to create the plot. It is
            recommended to call this manually, if you decide to save
            the plots yourself or when you want only the fit results.

            Parameters
            ----------
            fname : str, optional
                Path where to save, without file extension. Defaults to "./mre"
        """
        if not isinstance(fname, str):
            fname = str(fname)
        if fname == "":
            fname = "./mre"

        # try creating enclosing dir if not existing
        tempdir = os.path.abspath(os.path.expanduser(fname + "/../"))
        os.makedirs(tempdir, exist_ok=True)

        fname = os.path.expanduser(fname)

        log.info("Saving meta to {}.tsv".format(fname))
        # fits
        hdr = "Mr. Estimator v{}\n".format(__version__)
        try:
            for fdx, fit in enumerate(self.fits):
                hdr += "{}\n".format("-" * 72)
                hdr += "legendlabel: " + str(self.fitlabels[fdx]) + "\n"
                hdr += "{}\n".format("-" * 72)
                if fit.desc != "":
                    hdr += "description: " + str(fit.desc) + "\n"
                hdr += "m = {}\ntau = {} [{}]\n".format(fit.mre, fit.tau, fit.dtunit)
                if fit.quantiles is not None:
                    hdr += "quantiles | tau [{}] | m:\n".format(fit.dtunit)
                    for i, q in enumerate(fit.quantiles):
                        hdr += "{:6.3f} | ".format(fit.quantiles[i])
                        hdr += "{:8.3f} | ".format(fit.tauquantiles[i])
                        hdr += "{:8.8f}\n".format(fit.mrequantiles[i])
                    hdr += "\n"
                hdr += "fitrange: {} <= k <= {} [{} {}]\n".format(
                    fit.steps[0], fit.steps[-1], ut._printeger(fit.dt), fit.dtunit
                )
                hdr += "function: " + ut.math_from_doc(fit.fitfunc) + "\n"
                # hdr += '\twith parameters:\n'
                parname = list(inspect.signature(fit.fitfunc).parameters)[1:]
                parlen = len(max(parname, key=len))
                for pdx, par in enumerate(self.fits[fdx].popt):
                    unit = ""
                    if parname[pdx] == "nu":
                        unit += "[1/{}]".format(fit.dtunit)
                    elif parname[pdx].find("tau") != -1:
                        unit += "[{}]".format(fit.dtunit)
                    hdr += "\t{: <{width}}".format(
                        parname[pdx] + " " + unit, width=parlen + 5 + len(fit.dtunit)
                    )
                    hdr += " = {}\n".format(par)
                hdr += "\n"
        except Exception as e:
            log.debug("Exception passed", exc_info=True)

        # rks / ts
        labels = ""
        dat = []
        if self.ydata is not None and len(self.ydata) != 0:
            hdr += "{}\n".format("-" * 72)
            hdr += "Data\n"
            hdr += "{}\n".format("-" * 72)
            labels += "1_" + self.xlabel
            for ldx, label in enumerate(self.ylabels):
                labels += "\t" + str(ldx + 2) + "_" + label
            labels = labels.replace(" ", "_")
            dat = np.vstack((self.xdata, np.asarray(self.ydata)))
        np.savetxt(
            fname + ".tsv", np.transpose(dat), delimiter="\t", header=hdr + labels
        )


def overview(src, rks, fits, **kwargs):
    """
        creates an A4 overview panel and returns the matplotlib figure element.
        No Argument checks are done
    """

    ratios = np.ones(5)
    ratios[4] = 0.0001
    # ratios=None
    # A5 in inches, should check rc params in the future
    # matplotlib changes the figure size when modifying subplots
    fig, axes = plt.subplots(
        nrows=5, figsize=(5.8, 8.3), gridspec_kw={"height_ratios": ratios}
    )

    # avoid huge file size for many trials due to separate layers.
    # everything below 0 gets rastered to the same layer.
    axes[0].set_rasterization_zorder(0)

    # ------------------------------------------------------------------ #
    # Time Series
    # ------------------------------------------------------------------ #

    tsout = OutputHandler(ax=axes[0])
    tsout.add_ts(src, label="Trials")
    if src.shape[0] > 1:
        try:
            prevclr = plt.rcParams["axes.prop_cycle"].by_key()["color"][0]
        except Exception:
            prevclr = "navy"
            log.debug("Exception getting color cycle", exc_info=True)
        tsout.add_ts(np.mean(src, axis=0), color=prevclr, label="Average")
    else:
        tsout.ax.legend().set_visible(False)

    tsout.ax.set_title("Time Series", fontweight="bold", loc="center")
    tsout.ax.set_title("(Input Data)", fontsize="medium", color="#646464", loc="right")
    tsout.ax.set_xlabel(
        "t [{}{}]".format(
            ut._printeger(rks[0].dt) + " " if rks[0].dt != 1 else "", rks[0].dtunit
        )
    )

    # ------------------------------------------------------------------ #
    # Mean Trial Activity
    # ------------------------------------------------------------------ #

    if src.shape[0] > 1:
        # average trial activites as function of trial number
        taout = OutputHandler(rks[0].trialactivities, ax=axes[1])
        try:
            err1 = rks[0].trialactivities - np.sqrt(rks[0].trialvariances)
            err2 = rks[0].trialactivities + np.sqrt(rks[0].trialvariances)
            prevclr = plt.rcParams["axes.prop_cycle"].by_key()["color"][0]
            taout.ax.fill_between(
                np.arange(1, rks[0].numtrials + 1), err1, err2, color=prevclr, alpha=0.2
            )
        except Exception as e:
            log.debug("Exception adding std deviation to plot", exc_info=True)
        taout.ax.set_title("Mean Trial Activity and Std. Deviation", fontweight="bold")
        taout.ax.set_xlabel("Trial i")
        taout.ax.set_ylabel("$\\bar{A}_i$")
    else:
        # running average over the one trial to see if stays stationary
        numsegs = kwargs.get(numsegs) if "numsegs" in kwargs else 50
        ravg = np.zeros(numsegs)
        err1 = np.zeros(numsegs)
        err2 = np.zeros(numsegs)
        seglen = int(src.shape[1] / numsegs)
        for s in range(numsegs):
            temp = np.mean(src[0][s * seglen : (s + 1) * seglen])
            ravg[s] = temp
            stddev = np.sqrt(np.var(src[0][s * seglen : (s + 1) * seglen]))
            err1[s] = temp - stddev
            err2[s] = temp + stddev

        taout = OutputHandler(ravg, ax=axes[1])
        try:
            prevclr = plt.rcParams["axes.prop_cycle"].by_key()["color"][0]
            taout.ax.fill_between(
                np.arange(1, numsegs + 1), err1, err2, color=prevclr, alpha=0.2
            )
        except Exception as e:
            log.debug("Exception adding std deviation to plot", exc_info=True)
        taout.ax.set_title(
            "Average Activity and Stddev for {} Intervals".format(numsegs),
            fontweight="bold",
        )
        taout.ax.set_xlabel("Interval i")
        taout.ax.set_ylabel("$\\bar{A}_i$")

    # ------------------------------------------------------------------ #
    # Coefficients and Fit results
    # ------------------------------------------------------------------ #

    cout = OutputHandler(rks + fits, ax=axes[2])

    fitcurves = []
    fitlabels = []
    for i, f in enumerate(cout.fits):
        fitcurves.append(cout.fitcurves[i][0])
        label = ut.math_from_doc(f.fitfunc, 5)
        label += "\n\n$\\tau={:.2f}${}\n".format(f.tau, f.dtunit)
        if f.tauquantiles is not None:
            label += "$[{:.2f}:{:.2f}]$\n\n".format(
                f.tauquantiles[0], f.tauquantiles[-1]
            )
        else:
            label += "\n\n"
        label += "$m={:.5f}$\n".format(f.mre)
        if f.mrequantiles is not None:
            label += "$[{:.5f}:{:.5f}]$".format(f.mrequantiles[0], f.mrequantiles[-1])
        else:
            label += "\n"
        fitlabels.append(label)

    tempkwargs = {
        # 'title': 'Fitresults',
        "ncol": len(fitlabels),
        "loc": "upper center",
        "mode": "expand",
        "frameon": True,
        "markerfirst": True,
        "fancybox": False,
        # 'framealpha': 1,
        "borderaxespad": 0,
        "edgecolor": "black",
        # hide handles
        "handlelength": 0,
        "handletextpad": 0,
    }
    try:
        axes[3].legend(fitcurves, fitlabels, **tempkwargs)
    except Exception:
        log.debug("Exception passed", exc_info=True)
        del tempkwargs["edgecolor"]
        axes[3].legend(fitcurves, fitlabels, **tempkwargs)

    # hide handles
    for handle in axes[3].get_legend().legendHandles:
        handle.set_visible(False)

    # center text
    for t in axes[3].get_legend().texts:
        t.set_multialignment("center")

    # apply stile and fill legend
    axes[3].get_legend().get_frame().set_linewidth(0.5)
    axes[3].axis("off")
    axes[3].set_title(
        "Fitresults", fontweight="bold", loc="center",
    )
    axes[3].set_title(
        " (with CI: [$12.5\\%:87.5\\%$])",
        color="#646464",
        fontsize="medium",
        loc="right",
    )
    for a in axes:
        a.xaxis.set_tick_params(width=0.5)
        a.yaxis.set_tick_params(width=0.5)
        for s in a.spines:
            a.spines[s].set_linewidth(0.5)

    # dummy axes for version and warnings
    axes[4].axis("off")

    fig.tight_layout()
    plt.subplots_adjust(hspace=0.8, top=0.95, bottom=0.0, left=0.1, right=0.99)

    title = kwargs.get("title") if "title" in kwargs else None
    if title is not None and title != "":
        fig.suptitle(title, fontsize=14)
        plt.subplots_adjust(top=0.91)

    if "warning" in kwargs and kwargs.get("warning") is not None:
        s = "\u26A0 {}".format(kwargs.get("warning"))
        fig.text(0.5, 0.01, s, fontsize=13, horizontalalignment="center", color="red")

    fig.text(
        0.995,
        0.005,
        "v{}".format(__version__),
        fontsize="small",
        horizontalalignment="right",
        color="#646464",
    )

    return fig



import os
import numpy as np
import matplotlib.pyplot as plt

# import mre last to use (possibly) specified pyplot backend
import mrestimator as mre

# ------------------------------------------------------------------ #
# importing data
# ------------------------------------------------------------------ #

# set directory to the location of this script file to use relative paths
os.chdir(os.path.dirname(__file__))

# create an output directory to save results
os.makedirs('./output', exist_ok=True)

filepath = './data/full.tsv'

# import all columns of the file, each column is a trial
srcful = mre.input_handler(filepath)
print('srcful has shape: ', srcful.shape)

# import selected columns of a file, here the second column
srcdrv = mre.input_handler('./data/drive.tsv', usecols=1)
print('drive has shape: ', srcdrv.shape)

# or pass a list of files to import, every column of each file becomes a trial
filelist = [
    './data/sub_01.tsv',
    './data/sub_02.tsv',
    './data/sub_03.tsv']
srcsub = mre.input_handler(filelist)
print('imported trials from list: ', srcsub.shape[0])

# alternatively, you can use a wildcard to match the file pattern
srcsub = mre.input_handler('./data/sub_*.tsv')
print('imported trials from wildcard: ', srcsub.shape[0])

# use np.mean along axis 0 to get the average activity across all trials
# of the trial structure
avgful = np.mean(srcful, axis=0)
avgsub = np.mean(srcsub, axis=0)

# ------------------------------------------------------------------ #
# use the wrapper function to do all needed steps in the right order
# ------------------------------------------------------------------ #

# this function will change in the next weeks until we decide on a
# final interface
auto = mre.full_analysis(
    data='./data/sub_*.tsv',
    targetdir='./output',
    title='Full Analysis',
    dt=4, dtunit='ms',
    tmin=0, tmax=8000,
    fitfuncs=['exp', 'exp_offs', 'complex'],
    numboot='auto'
    )

plt.show()

# ------------------------------------------------------------------ #
# plotting time series
# ------------------------------------------------------------------ #

# create a new OutputHandler to add content with custom options
oful = mre.OutputHandler()

# add time series of the inputdata. if more than one trial, timeseries are
# plotted slightly transparent per default
oful.add_ts(srcful)

# keyword arguments "kwargs" are passed through to matplotlib,
# e.g. to specify a color or the label for the plot legend
oful.add_ts(avgful, color='navy', label='average (full)')

# colors improved a lot in newer versions of matplotlib.
# here we used web colors for compatiblity check the links below
# https://matplotlib.org/1.5.3/api/colors_api.html
# https://matplotlib.org/2.2.3/api/colors_api.html
# especially the color='C0' ... color='CN' notation is nice for accessing
# default colors in sequential order

# see https://matplotlib.org/api/_as_gen/matplotlib.pyplot.plot.html
# for some more style options
oful.add_ts(srcsub, alpha=0.25, color='yellow', label='trials (subs.)')
oful.add_ts(avgsub, ls='dashed', color='maroon', label='average (subs.)')

# add the drive
oful.add_ts(srcdrv, color='green', label='drive')

plt.show()

# ------------------------------------------------------------------ #
# analyzing
# ------------------------------------------------------------------ #

# correlation coefficients with default settings, assumes 1ms time bins
rkdefault = mre.coefficients(srcful)
print(rkdefault)
print('this guy has the following attributes: ', rkdefault._fields)

# specify the range of time steps (from, to) for which coefficients are wanted
# also, set the unit and the number of time steps per bin e.g. 4ms per k:
rk = mre.coefficients(srcsub, steps=(1, 5000), dt=4, dtunit='ms', desc='mydat')

# fit with defaults: exponential over the full range of rk
m = mre.fit(rk)

# specify a custom fit range and fitfunction.
m2 = mre.fit(rk, steps=(1, 3000), fitfunc='offset')
# you could also provide an np array containing all the steps you want to
# use, e.g. with strides other than one
# m2 = mre.fit(rk, steps=np.arange(1, 3000, 100), fitfunc='offset')

# Plot with a new handler, you can provide multiple things to add
ores = mre.OutputHandler([rkdefault, m])
# or add them individually, later
ores.add_coefficients(rk)
ores.add_fit(m2)
# Note the different time scales
# The description provided to mre.coefficients is automatically used for
# subsequent steps and becomes the axis label

# save the plot and its meta data
ores.save('./output/custom')

# show all figures and halt script until closed via gui
plt.show(block=True)





# ------------------------------------------------------------------ #
# Example how to use the mrestimator toolbox (v0.1.6)
# - on experimental spike data
# - how to split long time series into artificial trials
# - (in order to) obtain error estimates via bootstrapping
#
# The data by Mizuseki K, Sirota A, Pastalkova E, Buzsáki G. (2009)
# is freely available on crcns: http://dx.doi.org/10.6080/K0Z60KZ9
#
# "The data set contains multichannel simultaneous recordings made
# from layer CA1 of the right dorsal hippocampus of three Long-Evans
# rats during open field tasks"
#
# download the .tar.gz from
# https://portal.nersc.gov/project/crcns/download/hc-2/ec013.527
# and extract it into the same folder as this script
# ------------------------------------------------------------------ #

import os
import numpy as np
import matplotlib as mpl
import sys
import mrestimator as mre

# helper function to convert a list of time stamps
# into a (binned) time series of activity
def bin_spike_times_unitless(spike_times, bin_size):
    last_spike = spike_times[-1]
    num_bins = int(np.ceil(last_spike/bin_size))
    res = np.zeros(num_bins)
    for spike_time in spike_times:
        target_bin = int(np.floor(spike_time/bin_size))
        res[target_bin] = res[target_bin] + 1
    return res

# set directory to the location of this script file to use relative paths
os.chdir(os.path.dirname(os.path.abspath(__file__)))

# load the spiketimes
# res = np.loadtxt('./crcns/hc2/ec014.277/ec014.277.res.3') # ~8Hz oscillations
res = np.loadtxt('./crcns/hc2/ec013.527/ec013.527.res.1') # ~6Hz oscillations

# the .res.x files contain the time-stamps of spikes detected by electrode x
# sampled at 20kHz, i.e. 0.05ms per time steps.
# we want 'spiking activity', i.e. spikes added up during a given time window.
# usually, ~4ms time bins (windows) is a good initial guess
act = bin_spike_times_unitless(res, bin_size=80)

# to run the analysis on our activity time series with 4 ms time bins
# mre.full_analysis(act, dt=4, dtunit='ms', kmax=800)

# create 25 artifical trials by splitting the data to get error estimates
triallen = int(np.floor(len(act)/25))
trials = np.zeros(shape=(25, triallen))
for i in range(0, 25):
    trials[i] = act[i*triallen : (i+1)*triallen]

# now we could run the analysis and will get error estimates
# mre.full_analysis(trials, dt=4, dtunit='ms', kmax=800)

# however, in this dataset we will find theta oscillations.
# let's try the other fit functions, too.

# if we want error estimates for the complex fit, we can provide
# ```numboot=200```, but it is quite slow.

# to save the files to disk, we can specify a targetdirectory.
# in the text file, the frequency of the fitted oscillations can be found.

out = mre.full_analysis(trials, dt=4, dtunit='ms', kmax=800,
    method='trialseparated',
    fitfuncs=['exponential', 'exponential_offset', 'complex'],
    targetdir='./', saveoverview=True)

# by assigning the result of mre.full_analysis(...) to a variable, we can
# use fit results for further processing.

# the oscillation frequency nu is fitted by the complex fit functions as the
# 7th parameter. it is in units of 1/dtunit and we used 'ms'.
print(f"oscillation frequency: {out.fits[2].popt[6]*1000:.2f} [Hz]")






import numpy as np
import os
import matplotlib.pyplot as plt
from scipy.ndimage.filters import gaussian_filter1d
import scipy.optimize
import unittest
import pickle
import time
print(__name__)


#import mrestimator as mre
from mrestimator.utility import log
import mrestimator as mre

#Currently not working...

######################

def calc_popt(fitfunc, k_arr, corr_arr, weights, bounds, startvalues, maxfev=None, try_only_once=False):
    k_arr = np.array(k_arr, dtype="float64")
    corr_arr = np.array(corr_arr, dtype="float64")
    weights = np.array(weights, dtype="float64")
    #weights = gaussian_filter1d(weights, sigma = 10)
    bounds = np.array(bounds, dtype="float64")
    chisq_best = 1e42
    popt_best = None

    chisq_arr = []

    for startvalue in startvalues:

        startvalue = np.array(startvalue, dtype="float32")


        for val, (bound_min, bound_max), i in zip(startvalue, bounds, range(100000)):
            if val < bound_min or val > bound_max:
                print(val)
                raise (RuntimeError("startvalue {} outside bounds, {}-th value".format(val, i + 1)))


        if not np.any(np.isfinite(corr_arr)):
            raise RuntimeError("corr arr is not finite: {}".format(corr_arr))
        try:
            popt, pcov = scipy.optimize.curve_fit(fitfunc, xdata=k_arr, ydata=corr_arr, sigma=weights,
                                                  p0=startvalue,
                                                  bounds=np.array(list(zip(*bounds)), dtype="float32"), maxfev=maxfev,
                                                  ftol=2e-5, gtol=1e-5, xtol=1e-5, diff_step=2e-6, method="trf")
        except RuntimeError:
            continue
        except ValueError as err:
            print(startvalue)
            raise RuntimeError("{} : {}, {}".format(err, str(startvalue), bounds))
        r = corr_arr - fitfunc(k_arr, *popt)
        chisq = np.sum((r / weights) ** 2)
        chisq_arr.append(chisq)

        if chisq < chisq_best:
            popt_best = popt
            chisq_best = chisq

    #if none of the startvalues worked, try with more iterations
    if popt_best is None and not try_only_once:
        popt_best, chisq_best = calc_popt(fitfunc, k_arr, corr_arr, weights, bounds, startvalues, maxfev=6000,
                                          try_only_once=True)
    if popt_best is None and not try_only_once:
        popt_best, chisq_best = calc_popt(fitfunc, k_arr, corr_arr, weights, bounds, startvalues, maxfev=20000,
                                          try_only_once=True)

    return popt_best, chisq_best




def fitfunction_complex(k_arr, A, tau1, exponent, freq, O, B, tau2, C, tau3):
    return A * np.exp(-(np.abs(k_arr) / tau1) ** exponent) * np.cos(2 * np.pi * freq * k_arr) + O * np.ones_like(
        k_arr) + \
           B * np.exp(-(k_arr / tau2)) + C * np.exp(-(k_arr / tau3) ** 2)

def fitfunction_exp_with_offset(k_arr, B, tau2, O):
    return B * np.exp(-(k_arr / tau2)) + O * np.ones_like(k_arr)



def compare_mre_methods(activity_mat):
    print('comparing mre internally:')
    rk1 = mre.correlation_coefficients(activity_mat, maxstep=1500,
                                       method='trialseparated')
    rk2 = mre.correlation_coefficients(activity_mat, maxstep=1500,
                                       method='trialseparatedfit')
    rk3 = mre.correlation_coefficients(activity_mat, maxstep=1500,
                                       method='sationarymean')
    rk4 = mre.correlation_coefficients(activity_mat, maxstep=1500,
                                       method='sationarymeanfit')

    print('\n\n\n comparison:')
    print(rk1.coefficients[:10])
    print(rk2.coefficients[:10])
    print(rk3.coefficients[:10])
    print(rk4.coefficients[:10])
    print('\n\n\n')

    test_similarity(rk1.coefficients, rk2.coefficients)
    test_similarity(rk3.coefficients, rk4.coefficients)

class TestMREstimator(unittest.TestCase):

    startvalues_compl = [(0.03, 300, 1, 1. / 200, 0, 0.1, 10, 0.03, 10),
                         (0.03, 200, 2.5, 1. / 250, 0, 0.1, 400, 0.03, 25),
                         (0.03, 100, 1.5, 1. / 50, 0.03, 0.1, 20, 0.03, 10),
                         (0.03, 100, 1.5, 1. / 50, 0.03, 0.1, 300, 0.03, 10),
                         (0.03, 100, 1, 1 / 150, 0.01, 0.03, 20, 0.03, 5),
                         (0.03, 100, 1, 1 / 150, 0.01, 0.03, 20, 0.03, 5),
                         (0.03, 300, 1.5, 1 / 100, 0.03, 0.05, 10, 0.1, 5),
                         (0.03, 300, 1.5, 1 / 100, 0.03, 0.05, 300, 0.1, 10),
                         (0.010, 116, 2, 1 / 466, 0.010, 0.029, 56, 0.03, 5),
                         (0.010, 116, 2, 1 / 466, 0.010, 0.029, 56, 0.03, 5),
                         (0.010, 116, 2, 1 / 466, 0.010, 0.029, 56, 0.03, 5),
                         (0.017, 107, 1, 1 / 478, 0.044, 0.078, 19, 0.1, 5),
                         (0.017, 107, 1, 1 / 478, 0.044, 0.078, 19, 0.1, 5),
                         (0.067, 300, 2, 1 / 127, 0.045, 0.029, 10, 0.03, 10),
                         (0.03, 50, 1, 1 / 150, 0.012, 0.029, 210, 0.1, 10),
                         (0.03, 50, 1, 1 / 150, 0.012, 0.029, 210, 0.1, 10),
                         (0.03, 50, 1, 1 / 150, 0.012, 0.029, 210, 0.03, 10),
                         (0.03, 50, 1, 1 / 150, 0.012, 0.029, 210, 0.03, 10),
                         (0.08, 50, 1, 1 / 34, 0.002, 0.029, 310, 0.03, 5),
                         (0.08, 50, 1, 1 / 34, 0.002, 0.029, 310, 0.03, 5),
                         (0.08, 50, 1, 1 / 64, 0.002, 0.029, 310, 0.03, 5),
                         (0.08, 50, 1, 1 / 64, 0.002, 0.029, 310, 0.03, 5)]
    bounds_compl = [(-5, 5), (5, 5000), (1 / 3, 3), (2. / 1000, 50. / 1000), (-1, 1), (0, 1), (5, 5000),
                    (-5, 5), (0, 30)]
    startvalues_exp_with_offset = [(0.1, 100, 0)]
    bounds_exp_with_offset = [(0,1), (1,10000), (-1, 1)]

    def test_stationary_mean(self):
        for ele_num in range(0,40,10):
            for fitfunc_name in ["exponential_with_bias", "complex"]:
                name_data = "./data/activity_mat_{}.pickled".format(ele_num)
                with self.subTest(fitfunc_name = fitfunc_name, data_file = name_data):
                    k_arr = np.arange(7, 1500, 1)
                    activity_mat = pickle.load(open(name_data, "rb"))
                    corr_arr = calc_corr_arr_stationary(activity_mat, k_arr)
                    if fitfunc_name == "complex":
                        popt,_ = calc_popt(fitfunction_complex, k_arr, corr_arr, np.ones_like(k_arr), self.bounds_compl, self.startvalues_compl, maxfev=None,
                                  try_only_once=False)

                        rk = mre.coefficients(activity_mat, method='stationarymean')
                        res_mre = mre.correlation_fit(rk, fitfunc='complex')
                        print('popts:', popt, res_mre.popt)
                        for i in range(len(popt)):
                            self.assertTrue(test_similarity(popt[i], res_mre.popt[i], ratio_different=1e-5))
                        #plt.plot(k_arr, corr_arr)
                        #plt.plot(k_arr, fitfunction_complex(k_arr, *popt))
                        #plt.show()

                    elif fitfunc_name == "exponential_with_bias":
                        popt,_ = calc_popt(fitfunction_exp_with_offset, k_arr, corr_arr, np.ones_like(k_arr), self.bounds_exp_with_offset,
                                         self.startvalues_exp_with_offset, maxfev=None,
                                         try_only_once=False)
                        #plt.plot(k_arr, corr_arr)
                        #plt.plot(k_arr, fitfunction_exp_with_offset(k_arr, *popt))
                        #plt.show()
                        rk = mre.coefficients(activity_mat, method='stationarymean')
                        res_mre = mre.correlation_fit(rk, fitfunc='exponentialoffset')
                        print('popts:', popt, res_mre.popt)
                        for i in range(len(popt)):
                            self.assertTrue(test_similarity(popt[i], res_mre.popt[i], ratio_different=1e-5))







import unittest
import pickle
import time
import sys

import numpy as np

import mrestimator as mre
from mrestimator.utility import log


def test_similarity(value1, value2, ratio_different=1e-10):
    print('ratio difference: {:.3e}'.format(np.max(np.fabs(value1 - value2)/((value1 + value2)/2))))
    return np.all(np.fabs(value1 - value2)/((value1 + value2)/2)  < ratio_different)

def test_similarity_abs(value1, value2, max_difference=1e-10):
    print('max difference: {:.3e}'.format(np.max(np.fabs(value1 - value2))))
    return np.all(np.fabs(value1 - value2)  < max_difference)

def calc_corr_arr_stationary(activity_mat, k_arr):
    average = np.mean(activity_mat)
    corr_arr = np.zeros_like(k_arr, dtype="float64")
    n = len(activity_mat[0])
    variance = np.mean((activity_mat[:]-average)**2)*(n/(n-1))
    for i, k in enumerate(k_arr):
        corr_arr[i] = np.mean((activity_mat[:,:-k]-average) * (activity_mat[:,k:] - average)) * ((n-k)/(n-k-1)) / variance
    return corr_arr


def calc_corr_arr_stationary_new(activity_mat, k_arr):
    corr_arr = np.zeros((len(k_arr)), dtype="float64")
    for i, k in enumerate(k_arr):
        x = activity_mat[:,:-k]
        y = activity_mat[:,k:]
        x_mean = np.mean(x)
        y_mean = np.mean(y)
        corr_arr[i] = np.sum(np.mean((x-x_mean) * (y - y_mean), axis=1))/\
                      np.sum(np.mean((x-x_mean)**2, axis=1))
    return corr_arr

def calc_corr_mat_separate(activity_mat, k_arr):
    corr_arr = np.zeros((len(k_arr)), dtype="float64")
    for i, k in enumerate(k_arr):
        x = activity_mat[:,:-k]
        y = activity_mat[:,k:]
        x_mean = np.mean(x, axis=1)[:, np.newaxis]
        y_mean = np.mean(y, axis=1)[:, np.newaxis]
        corr_arr[i] = np.mean(np.mean((x-x_mean) * (y - y_mean), axis=1)/
                              np.var(x, axis=1))
    return corr_arr

class TestCorrCoeff(unittest.TestCase):
    log.setLevel(40)

    def test_stationary_mean(self):
        print("\nTesting stationary mean correlation coefficients: \n")

        for ele_num in range(0,40,10):
            name_data = "./data/activity_mat_{}.pickled".format(ele_num)
            activity_mat = pickle.load(open(name_data, "rb"))
            k_arr = np.arange(7, 1500, 1)
            activity_mat = activity_mat.astype(dtype="float64")
            corr_arr = calc_corr_arr_stationary_new(activity_mat, k_arr)
            time_beg = time.time()

            numboot = 100

            mre_res = mre.coefficients(activity_mat,
                             steps=k_arr,
                             method='stationarymean',
                             numboot=numboot)
            print('stationarymean, time needed:  {:.2f} ms'.format((time.time()-time_beg)*1000))
            print('mre: ', mre_res.coefficients[:5])
            print('true value: ', corr_arr[:5])

            self.assertTrue(test_similarity(mre_res.coefficients, corr_arr, ratio_different = 1e-8))
            bootstrap_mat = np.array([boot.coefficients for boot in mre_res.bootstrapcrs])
            mean_bootstrap = np.mean(bootstrap_mat, axis=0)
            print("boot mean: ", mean_bootstrap[:5])
            self.assertTrue(test_similarity_abs(mre_res.coefficients, np.mean(bootstrap_mat, axis=0),
                                                max_difference=0.04/np.sqrt(numboot)))

    def test_separate(self):
        print("\nTesting trial separated correlation coefficients: \n")


        for ele_num in range(0,40,10):
            name_data = "./data/activity_mat_{}.pickled".format(ele_num)
            activity_mat = pickle.load(open(name_data, "rb"))
            activity_mat = activity_mat.astype(dtype="float64")
            k_arr = np.arange(7, 1500, 1)
            corr_arr = calc_corr_mat_separate(activity_mat, k_arr)
            time_beg = time.time()
            numboot = 100
            mre_res = mre.coefficients(activity_mat,
                             steps=k_arr,
                             method='trialseparated',
                             numboot=numboot)

            print('trialseparated, time needed:  {:.2f} ms'.format((time.time()-time_beg)*1000))
            print('mre: ', mre_res.coefficients[:5])
            print('true value: ', corr_arr[:5])

            self.assertTrue(test_similarity(mre_res.coefficients, corr_arr, ratio_different = 1e-10))
            bootstrap_mat = np.array([boot.coefficients for boot in mre_res.bootstrapcrs])
            mean_bootstrap = np.mean(bootstrap_mat, axis=0)
            print("boot mean: ", mean_bootstrap[:5])
            self.assertTrue(test_similarity_abs(mre_res.coefficients, np.mean(bootstrap_mat, axis=0),
                                                max_difference=0.04/np.sqrt(numboot)))


from mrestimator.coefficients import *
from mrestimator.simulate import simulate_branching

class TestCCKnownMean(unittest.TestCase):
    def test_sm(self):
        print("Testing knownmean argument to sm_method")
        data  = simulate_branching(m=.98, a=100, numtrials=5)
        steps = np.arange(1,25,1)
        mx, my, x_y, x_x = sm_precompute(data, steps, 0.0)
        assert(np.all(mx == 0))
        assert(np.all(my == 0))

        mx, my, x_y, x_x = sm_precompute(data, steps, 5.0)
        assert(np.all(mx == 5))
        assert(np.all(my == 5))

        sm_prepped = mx, my, x_y, x_x
        rk = sm_method(sm_prepped, steps)
        rk2 = coefficients(data, steps=steps, method='sm',
            knownmean = 5.0).coefficients
        assert(test_similarity(rk, rk2, ratio_different = 1e-8))

        mx2, my2, x_y2, x_x2 = sm_precompute(data, steps, None)
        assert(mx2.shape  == mx.shape)
        assert(my2.shape  == my.shape)
        assert(x_y2.shape == x_y.shape)
        assert(x_x2.shape == x_x.shape)

        sm_prepped = mx2, my2, x_y2, x_x2
        rk = sm_method(sm_prepped, steps)
        # check that this matches the default.
        rk2 = coefficients(data, steps=steps, method='sm').coefficients
        assert(test_similarity(rk, rk2, ratio_different = 1e-8))

    def test_ts(self):
        print("Testing knownmean argument to ts_method")
        data  = simulate_branching(m=.98, a=100, numtrials=5)
        steps = np.arange(1,25,1)
        ts_prepped = ts_precompute(data, steps, 0.0)
        rk = ts_method(ts_prepped, steps)
        rk2 = coefficients(data, steps=steps, method='ts',
            knownmean=0.0).coefficients
        # nothing we can access here to compare...

        ts_prepped = ts_precompute(data, steps, None)
        rk = ts_method(ts_prepped, steps)
        rk2 = coefficients(data, steps=steps, method='ts').coefficients
        assert(test_similarity(rk, rk2, ratio_different = 1e-8))

if __name__ == "__main__":
    unittest.main()



import logging

import numpy as np
import scipy

from mrestimator import utility as ut
log = ut.log

def simulate_branching(
    m,
    a=None,
    h=None,
    length=10000,
    numtrials=1,
    subp=1,
    seed='random'):
    """
        Simulates a branching process with Poisson input. Returns data
        in the trial structure.

        Per default, the function discards the first
        few time steps to produce stationary activity. If a
        `drive` is passed as ``h=0``, the recording starts instantly
        (and produces exponentially decaying activity).

        Parameters
        ----------
        m : float
            Branching parameter.

        a : float
            Stationarity activity of the process.
            Only considered if no drive `h` is specified.

        h : ~numpy.array, optional
            Specify a custom drive (possibly changing) for every time step.
            If `h` is given, its length takes priority over the `length`
            parameter. If the first or only value of `h` is zero, the recording
            starts instantly with set activity `a` and the resulting timeseries
            will not be stationary in the beginning.

        length : int, optional
            Number of steps for the process, thereby sets the total length of
            the generated time series. Overwritten if drive `h` is set as an
            array.

        numtrials : int, optional
            Generate 'numtrials' trials. Default is 1.

        seed : int, optional
            Initialise the random number generator with a seed. Per default,
            ``seed='random'`` and the generator is seeded randomly (hence
            each call to `simulate_branching()` returns different results).
            ``seed=None`` skips (re)seeding.

        subp : float, optional
            Subsample the activity with the probability `subp` (calls
            `simulate_subsampling()` before returning).

        Returns
        -------
        : :class:`~numpy.ndarray`
            with `numtrials` time series, each containging
            `length` entries of activity.
            Per default, one trial is created with
            10000 measurements.
    """

    length = int(length)
    numtrials = int(numtrials)
    if h is None:
        if a is None:
            log.exception("Missing argument, either provide " +
                "the activity 'a' or the drive 'h'")
            raise TypeError
        else:
            h = np.full((length), a * (1 - m))
    else:
        if a is None:
            a = 0
        h = np.asarray(h)
        if h.size == 1:
            h = np.full((length), h)
        elif len(h.shape) != 1:
            log.exception("Argument drive 'h' needs to be a float or 1d array")
            raise ValueError
        else:
            length = h.size

    log.debug('simulate_branching() seeding to {}'.format(seed))
    if seed is None:
        pass
    elif seed == 'random':
        np.random.seed(None)
    else:
        np.random.seed(seed)

    if h[0] == 0 and a != 0:
        log.debug('Skipping thermalization since initial h=0')
    if h[0] == 0 and a == 0:
        log.warning('activity a=0 and initial h=0')

    log.info('Generating branching process with m={}'.format(ut._printeger(m)))
    log.debug(
        '{:d} trials with {:d} time steps each\n'.format(numtrials, length) +
        'branchign ratio m={}\n'.format(m) +
        '(initial) activity a={}\n'.format(a) +
        '(initial) drive rate h={}'.format(h[0])
    )

    A_t = np.zeros(shape=(numtrials, length), dtype=int)
    a = np.ones_like(A_t[:, 0])*a

    # if drive is zero, user would expect exp-decay of set activity
    # for m>1 we want exp-increase, else
    # avoid nonstationarity by discarding some steps
    if (h[0] != 0 and h[0] and m < 1):
        therm = np.fmax(100, int(length*0.05))
        log.info('Setting up stationarity, {:d} steps'.format(therm))
        for idx in range(0, therm):
            a = np.random.poisson(lam=m*a + h[0])

    A_t[:, 0] = np.random.poisson(lam=m*a + h[0])
    for idx in range(1, length):
        try:
            # if m >= 1 activity may explode until this throws an error
            A_t[:, idx] = np.random.poisson(lam=m*A_t[:, idx-1] + h[idx])
        except ValueError as e:
            log.debug('Exception passed for bp generation', exc_info=True)
            # A_t.resize((numtrials, idx))
            A_t = A_t[:, 0:idx]
            log.info('Activity is exceeding numeric limits, canceling ' +
                'and resizing output from length={} to {}'.format(length, idx))
            break

    if subp != 1 and subp is not None:
        try:
            # do not change rng seed when calling this as nested, otherwise
            # bp with subs. is not reproducible even with given seed
            return simulate_subsampling(A_t, prob=subp, seed=None)
        except ValueError:
            log.debug('Exception passed', exc_info=True)
    return A_t

def simulate_subsampling(data, prob=0.1, seed='random'):
    """
        Apply binomial subsampling.

        Parameters
        ----------
        data : ~numpy.ndarray
            Data (in trial structre) to subsample. Note that `data` will be
            cast to integers. For instance, if your activity is normalised
            consider multiplying with a constant.

        prob : float
            Subsample to probability `prob`. Default is 0.1.

        seed : int, optional
            Initialise the random number generator with a seed. Per default set
            to `random`: seed randomly (hence each call to
            `simulate_branching()` returns different results).
            Set `seed=None` to keep the rng device state.
    """
    log.debug('simulate_subsampling()')
    if prob <= 0 or prob > 1:
        log.exception('Subsampling probability should be between 0 and 1')
        raise ValueError

    data = np.asarray(data)
    if len(data.shape) != 2:
        log.exception('Provide data as 2d ndarray (trial structure)')
        raise ValueError

    # activity = np.mean(data)
    # a_t = np.empty_like(data)

    log.debug('simulate_subsampling() seeding to {}'.format(seed))

    # we are always using the global random state device, although stats.binom
    # can have a local instance.
    if seed is None:
        pass
    elif seed == 'random':
        np.random.seed(None)
    else:
        np.random.seed(seed)

    # binomial subsampling, seed = None does not reseed global instance
    return scipy.stats.binom.rvs(data.astype(int), prob, size=data.shape)



import logging
from collections import namedtuple

import numpy as np

from mrestimator import utility as ut
log = ut.log
tqdm = ut.tqdm

# set precision of temporary results for numpy and numba
# ftype = np.longdouble # very slow, maybe float64 is enough
ftype=np.float64

try:
    from numba import jit, prange
    # raise ImportError
    log.info('Using numba for parallelizable functions')

    # implement needed sum functions to be compiled by numba:
    # parallelize higher level loops, during sm and ts methods
    @jit(nopython=True, parallel=False, fastmath=True, cache=True)
    def sum_1d(a):
        total = ftype(0)
        for i in prange(a.shape[0]):
            total += ftype(a[i])
        return total

    @jit(nopython=True, parallel=False, fastmath=True, cache=True)
    def sum_2d(a):
        total = ftype(0)
        for i in prange(a.shape[0]):
            for j in prange(a.shape[1]):
                total += ftype(a[i,j])
        return total

    @jit(nopython=True, parallel=False, fastmath=True, cache=True)
    def sum_2d_ax0(a):
        total = np.zeros((a.shape[1]), dtype=ftype)
        for i in prange(a.shape[0]):
            for j in prange(a.shape[1]):
                total[j] += ftype(a[i,j])
        return total

    @jit(nopython=True, parallel=False, fastmath=True, cache=True)
    def sum_2d_ax1(a):
        total = np.zeros((a.shape[0]), dtype=ftype)
        for i in prange(a.shape[0]):
            for j in prange(a.shape[1]):
                total[i] += ftype(a[i,j])
        return total

except ImportError:
    log.info('Numba not available, skipping parallelization')
    # replace numba functions if numba not available:
    # we only use jit and prange
    # helper needed for decorators with kwargs
    def parametrized(dec):
        def layer(*args, **kwargs):
            def repl(f):
                return dec(f, *args, **kwargs)
            return repl
        return layer

    @parametrized
    def jit(func, **kwargs):
        return func

    def prange(*args):
        return range(*args)

    def sum_1d(a):
        return np.sum(a, dtype=ftype)

    def sum_2d(a):
        return np.sum(a, dtype=ftype)

    def sum_2d_ax0(a):
        return np.sum(a, axis=0, dtype=ftype)

    def sum_2d_ax1(a):
        return np.sum(a, axis=1, dtype=ftype)


# ------------------------------------------------------------------ #
# Core routines for differnt coefficient methods
# ------------------------------------------------------------------ #

@jit(nopython=True, parallel=False, fastmath=True, cache=True)
def sm_precompute(data, steps, knownmean=None):
    """
        Part 1 of the >= v0.1.5 stationary mean method.
        Works for m>1
        Computes terms that are reused during bootstrapping.
        If knownmean is provided, we skip estimating the mean activity.
    """

    numsteps  = steps.shape[0]
    numtrials = data.shape[0]
    numels    = data.shape[1]

    # (x-mx)(y-my) = x*y + mx*my - my*x - mx*y
    x_y   = np.empty(shape=(numsteps, numtrials))
    x_x   = np.empty(shape=(numsteps, numtrials))
    if knownmean is None:
        mx    = np.empty(shape=(numsteps, numtrials))
        my    = np.empty(shape=(numsteps, numtrials))

        # precompute things that can be separated by trial and k
        mm     = sum_2d_ax1(data[:, :]   )
        mm_squ = sum_2d_ax1(data[:, :]**2)

        for idx in prange(len(steps)):
            k = steps[idx]
            x = data[:, 0:-k]
            y = data[:, k:  ]
            l = data[:, 0: k]
            r = data[:,-k:  ]
            x_y[idx] =           sum_2d_ax1(x*y) /(numels-k)
            x_x[idx] = (mm_squ - sum_2d_ax1(r*r))/(numels-k)
            mx [idx] = (mm     - sum_2d_ax1(r  ))/(numels-k)
            my [idx] = (mm     - sum_2d_ax1(l  ))/(numels-k)

    else:
        mx    = np.ones(shape=(numsteps, numtrials))*knownmean
        my    = np.ones(shape=(numsteps, numtrials))*knownmean

        mm_squ = sum_2d_ax1(data[:, :]**2)

        for idx in prange(len(steps)):
            k = steps[idx]
            x = data[:, 0:-k]
            y = data[:, k:  ]
            l = data[:, 0: k]
            r = data[:,-k:  ]
            x_y[idx] =           sum_2d_ax1(x*y) /(numels-k)
            x_x[idx] = (mm_squ - sum_2d_ax1(r*r))/(numels-k)

    return mx, my, x_y, x_x

@jit(nopython=True, parallel=False, fastmath=True, cache=True)
def sm_method(precomputed, steps, choices = None):
    """
        Part 2 of the >= v0.1.5 stationary mean method.
        Works for m>1
        Relies on the results from sm_percompute.
        Fun fact: `choices = ...` is equivalent to not specifying the index.
    """
    mx, my, x_y, x_x = precomputed

    if choices is None:
        x_y_ = x_y[:, :]
        x_x_ = x_x[:, :]
        mx_  = mx[:, :]
        my_  = my[:, :]
    else:
        x_y_ = x_y[:, choices]
        x_x_ = x_x[:, choices]
        mx_  = mx[:, choices]
        my_  = my[:, choices]

    res  = np.zeros(shape=(len(steps)), dtype=np.float64)
    norm  = len(mx[0])
    for idx in prange(len(steps)):
        mxk   = sum_1d(mx_[idx])/norm
        myk   = sum_1d(my_[idx])/norm
        y_mxk = my_[idx]*mxk
        x_myk = mx_[idx]*myk

        res[idx] = \
            (sum_1d(x_y_[idx] - x_myk - y_mxk)/norm + mxk*myk) \
            / (sum_1d(x_x_[idx]-2*mx_[idx]*mxk + mxk**2)/norm)

    return res

@jit(nopython=True, parallel=False, fastmath=True, cache=True)
def ts_precompute(data, steps, knownmean=None):
    """
        Part 1 of the trialseparated method.
        Containts the core of the method.
        For ts, precomputing is not needed, this is only for consistency with
        sm. Hence, ts_method only does one reduction based on the bootstrap
        trial choices.
        Also, this allows to implement knownmean.
    """
    N = data.shape[0]
    T = data.shape[1]
    res = np.zeros(shape=(N, len(steps)), dtype=np.float64)
    if knownmean is None:
        for idx in prange(len(steps)):
            k = steps[idx]
            frontmean = np.empty((N,1), ftype)
            backmean  = np.empty((N,1), ftype)
            frontmean[:,0] = sum_2d_ax1( data[:,  :-k]              )/(T-k)
            frontvar       = sum_2d_ax1((data[:,  :-k]-frontmean)**2)/(T-k)
            backmean[:,0]  = sum_2d_ax1( data[:, k:  ]              )/(T-k)

            res[:, idx] = \
                sum_2d_ax1((data[:, :-k] -frontmean)*(data[:, k:] -backmean)) \
                / frontvar / (T-k)
    else:
        themean = np.ones((N,1), ftype)*knownmean
        for idx in prange(len(steps)):
            k = steps[idx]
            frontvar = sum_2d_ax1((data[:,  :-k]-themean)**2)/(T-k)

            res[:, idx] = \
                sum_2d_ax1((data[:, :-k] -themean)*(data[:, k:] -themean))\
                / frontvar / (T-k)

    return res

@jit(nopython=True, parallel=False, fastmath=True, cache=True)
def ts_method(precomputed, steps, choices = None):
    """
        See ts_precompute.
    """
    if choices is None:
        res = sum_2d_ax0(precomputed)/precomputed.shape[0]
    else:
        res = sum_2d_ax0(precomputed[choices])/precomputed.shape[0]
    # res = np.mean(precomputed[choices], axis=0, dtype=ftype)
    return res

def sm_method_naive(data, steps):
    """
        Native version of stationary mean method.
        Not used, skips precomputing. Results *should* be the same as from
        sm_method()
    """
    numels = data.shape[1]
    res    = np.zeros(shape=len(steps), dtype='float64')
    for idx, k in enumerate(steps):
        # analogeous to trial separated
        frontmean = np.mean(data[:,  :-k], keepdims=True, dtype=ftype)
        frontvar  = np.var( data[:,  :-k], ddof=1,        dtype=ftype)
        backmean  = np.mean(data[:, k:  ], keepdims=True, dtype=ftype)

        res[idx] = \
            np.mean((data[:,  :-k] - frontmean) * \
                    (data[:, k:  ] - backmean ), dtype=ftype) \
            * ((numels-k)/(numels-k-1)) / frontvar

    return res

# ------------------------------------------------------------------ #
# Coefficient Result class
# ------------------------------------------------------------------ #

class CoefficientResult(namedtuple('CoefficientResultBase', [
    'coefficients',
    'steps',
    'dt',
    'dtunit',
    'method',
    'stderrs',
    'trialactivities',
    'trialvariances',
    'triallen',
    'bootstrapcrs',
    'trialcrs',
    'desc',
    'description',
    'numtrials',
    'numboot',
    'numsteps'])):
    """
        Result returned by `coefficients()`. Subclassed from
        :obj:`~collections.namedtuple`.

        Attributes are set to `None` if the specified method or input
        data do not provide them. All attributes of type :obj:`~numpy.ndarray`
        and lists are one-dimensional.

        Attributes
        ----------
        coefficients : ~numpy.ndarray
            Contains the coefficients :math:`r_k`, has length
            `numsteps`. Access via ``.coefficients[step]``

        steps : ~numpy.ndarray
            Array of the :math:`k` values matching `coefficients`.

        dt : float
            The size of each step in `dtunits`. Default is 1.

        dtunit : str
            Units of step size. Default is `'ms'`.

        method : str or None
            The method that was used to calculate the coefficients

        stderrs : ~numpy.ndarray or None
            Standard errors of the :math:`r_k`.

        trialactivities : ~numpy.ndarray
            Mean activity of each trial in the provided data.
            To get the global mean activity, use ``np.mean(trialactivities)``.
            Has lenght `numtrials`

        description : str
            Description (or name) of the data set, by default all results of
            functions working with this set inherit its description (e.g. plot
            legends).

        numtrials : int,
            Number of trials that contributed.

        numboot : int,
            Number of bootstrap replicas that were created.

        numsteps : int,
            Number of steps in `coefficients`, `steps` and `stderrs`.

        bootstrapcrs : list
            List containing the `numboot` :obj:`CoefficientResult` instances
            that were calculated from the resampled input data. The List is
            empty if bootstrapping was skipped (`numboot=0`).

        trialcrs : list
            List of the :obj:`CoefficientResult` instances calculated
            from individual trials. Only has length `numtrials` if the
            `trialseparated` method was used, otherwise it is empty.

        Note
        ----
        At the time of writing, :obj:`~numpy.ndarray` behaves a bit unexpected
        when creating arrays with objects that are sequence like (such as
        :obj:`CoefficientResult` and :obj:`FitResult`), even when specifying
        `dtype=object`.
        Numpy converts the objects into an ndimensional structure instead
        of creating the (probably desired) 1d-array. To work around the issue,
        use a `list` or manually create the array with `dtype=object` and add
        the entries after creation.

        Example
        -------
        .. code-block:: python

            import numpy as np
            import matplotlib.pyplot as plt
            import mrestimator as mre

            # branching process with 15 trials
            bp = mre.simulate_branching(m=0.995, a=10, numtrials=15)

            # the bp returns data already in the right format
            rk = mre.coefficients(bp, method='ts', dtunit='step')

            # fit
            ft = mre.fit(rk)

            # plot coefficients and the autocorrelation fit
            mre.OutputHandler([rk, ft])
            plt.show()

            # print the coefficients
            print(rk.coefficients)

            # get the documentation
            print(help(rk))

            # rk is inherited from namedtuple with all the bells and whistles
            print(rk._fields)
        ..
    """

    # prohibit adding attributes
    __slots__ = ()

    # custom constructor with default arguments and arg check
    def __new__(cls,
        coefficients,
        steps,
        dt              = 1.0,
        dtunit          = 'ms',
        method          = None,
        stderrs         = None,
        trialactivities = np.array([]),
        trialvariances  = np.array([]),
        triallen        = 0,
        bootstrapcrs    = np.array([]),
        trialcrs        = np.array([]),
        description     = None,
        desc            = None):

        # given attr check
        coefficients    = np.asarray(coefficients)
        steps           = np.asarray(steps)
        dt              = float(dt)
        dtunit          = str(dtunit)
        method          = None if method is None else str(method)
        stderrs         = None if stderrs is None else np.asarray(stderrs)
        trialactivities = np.asarray(trialactivities)
        trialvariances  = np.asarray(trialvariances)
        triallen        = int(triallen)
        bootstrapcrs    = bootstrapcrs if isinstance(bootstrapcrs, list) else \
            [bootstrapcrs]
        trialcrs        = trialcrs if isinstance(trialcrs, list) else \
            [trialcrs]
        description     = None if description is None else str(description)
        desc            = '' if description is None else str(description)

        # derived attr
        numtrials = len(trialactivities)
        numboot   = len(bootstrapcrs)
        numsteps  = len(coefficients)

        # order of args has to match above!
        return super(CoefficientResult, cls).__new__(cls,
            coefficients,
            steps,
            dt,
            dtunit,
            method,
            stderrs,
            trialactivities,
            trialvariances,
            triallen,
            bootstrapcrs,
            trialcrs,
            desc,
            description,
            numtrials,
            numboot,
            numsteps)

    # printed representation
    def __repr__(self):
        return '<%s.%s object at %s>' % (
        self.__class__.__module__,
        self.__class__.__name__,
        hex(id(self))
    )

    # used to compare instances in lists
    def __eq__(self, other):
        return self is other


 # for idx, k in enumerate(steps):


def coefficients(
    data,
    method=None,
    steps=None,
    dt=1, dtunit='ms',
    knownmean=None,
    numboot=100,
    seed=5330,
    description=None,
    desc=None,
    ):
    """
        Calculates the coefficients of correlation :math:`r_k`.

        Parameters
        ----------
        data : ~numpy.ndarray
            Input data, containing the time series of activity in the trial
            structure. If a one dimensional array is provieded instead, we
            assume a single trial and reshape the input.

        method : str
            The estimation method to use, either `'trialseparated'` (``'ts'``) or
            `'stationarymean'` (``'sm'``).
            ``'ts'`` calculates the :math:`r_k` for each trial separately and averages
            over trials. The resulting coefficients can be biased if the trials are
            too short.
            ``'sm'`` assumes the mean activity and its variance to be constant across
            all trials. The mean activity is then calculated from the larger pool
            of data from all trials and the short-trial-bias might be compensated.
            If you are unsure, compare results from both methods. If they agree,
            trials should be long enough.

        steps : ~numpy.array, optional
            Specify the steps :math:`k` for which to compute coefficients
            :math:`r_k`.
            If an array of length two is provided, e.g.
            ``steps=(minstep, maxstep)``, all enclosed integer values will be
            used.
            Arrays larger than two are assumed to contain a manual choice of
            steps. Strides other than one are possible.

        dt : float, optional
            The size of each step in `dtunits`. Default is 1.

        dtunit : str, optional
            Units of step size. Default is `'ms'`.

        description : str, optional
            Set the description of the :class:`CoefficientResult`. By default
            all results of functions working with this set inherit its
            description (e.g. plot legends).

        Other Parameters
        ----------------

        knownmean : float, optional
            If the (stationary) mean activity is known beforehand, it can be
            provided here. In this case, the provided value is used instead
            of approximating the expecation value of the activity using the
            mean.

        numboot : int, optional
            Enable bootstrapping to generate `numboot` (resampled)
            series of trials from the provided one. This allows to approximate
            statistical errors, returned in `stderrs`.
            Default is `numboot=100`.

        seed : int, None or 'random', optional
            If bootstrapping (`numboot>0`), a custom seed can be passed to
            the random number generator used for
            resampling. Per default, it is set to the *same* value every time
            `coefficients()` is called to return consistent results
            when repeating the analysis on the same data. Set to `None` to
            prevent (re)seeding. 'random' seeds using the wall clock time.
            For more details, see
            :obj:`numpy.random.RandomState`.

        Returns
        -------
        : :class:`CoefficientResult`
            The output is grouped and can be accessed
            using its attributes (listed below).
    """

    # ------------------------------------------------------------------ #
    # Check arguments to offer some more convenience
    # ------------------------------------------------------------------ #

    if desc is not None and description is None:
        description = str(desc);
    if description is not None:
        description = str(description)

    if knownmean is not None:
        knownmean = float(knownmean)

    # check dt
    dt = float(dt)
    if dt <= 0:
        log.exception('Timestep dt needs to be a float > 0.0')
        raise ValueError
    dtunit = str(dtunit)

    # check data
    dim = -1
    try:
        dim = len(data.shape)
        if dim == 1:
            log.warning('You should provide an ndarray of ' +
                'shape(numtrials, datalength)\n' +
                'Continuing with one trial, reshaping your input')
            data = np.reshape(data, (1, len(data)))
        elif dim >= 3:
            log.exception('Provided ndarray is of dim {}\n'.format(dim) +
                  'Please provide a two dimensional ndarray')
            raise ValueError
    except Exception as e:
        log.exception('Please provide a two dimensional ndarray')
        raise ValueError from e

    # check method
    log.debug('coefficients() using \'{}\' method:'.format(method))
    if method is None:
        if data.shape[0] == 1:
            method = 'sm'  # sm has a more useful warning when tau is large
            log.debug(
                "No coefficient method provided, using 'stationarymean' for one trial"
            )
        else:
            log.exception(
                "The provided data seems to have more than one trial. " +
                "Please specify a 'method':\n" +
                "'trialseparated' --- if your trials are long, or\n" +
                "'stationarymean' --- if you are sure that activity is stationary " +
                "across trials.\n" +
                "If you are unsure, we suggest to compare results from both methods."
            )
            raise ValueError

    if method not in ['trialseparated', 'ts', 'stationarymean', 'sm']:
        log.exception('Unknown method: "{}"'.format(method))
        raise NotImplementedError
    if method == 'ts':
        method = 'trialseparated'
    elif method == 'sm':
        method = 'stationarymean'

    # check steps
    if steps is None:
        steps = (None, None)
    try:
        steps = np.array(steps)
        assert len(steps.shape) == 1
    except Exception as e:
        log.exception('Please provide steps as ' +
            'steps=(minstep, maxstep) or as one dimensional numpy ' +
            'array containing all desired integer step values')
        raise ValueError from e
    if len(steps) == 2:
        minstep=1
        # default length not sure yet. but kmax > numels/2 is no use.
        maxstep=int(data.shape[1]/10)
        if steps[0] is not None:
            minstep = steps[0]
        if steps[1] is not None:
            maxstep = steps[1]
        if minstep > maxstep or minstep < 1:
            log.debug('minstep={} is invalid, setting to 1'.format(minstep))
            minstep = 1

        # it's important that kmax not larger than numels/2
        if maxstep > data.shape[1]/2 or maxstep < minstep:
            log.debug('maxstep={} is invalid'.format(maxstep))
            maxstep = int(data.shape[1]/2)
            log.debug('Adjusting maxstep to {}'.format(maxstep))
        steps = np.arange(minstep, maxstep+1, dtype=int)
        log.debug('Using steps between {} and {}'.format(minstep, maxstep))
    else:
        # dont overwrite provided argument
        steps = np.array(steps, dtype=int, copy=True)
        if (steps<1).any():
            log.warning(
                'All provided steps should be >= 1, modifying the input')
            incorrect = np.nonzero(steps < 1)
            correct = np.nonzero(steps >= 1)
            # np.arange(0,1000,10) -> only first element is a problem
            if (len(incorrect) == 1 and incorrect[0] == 0):
                if not (steps == 1).any():
                    steps[0] = 1
                    log.debug('Changed first element to 1')
                else:
                    steps = steps[1:]
                    log.debug('Removed first element')
            else:
                steps = steps[correct]
                log.debug('Only using steps that are >= 1')
        if (steps>data.shape[1]/2).any():
            log.warning('Provided steps include values that seem too large: ' +
                "Steps greater than half the time series length cause problems. " +
                "Proceeding nonetheless...")
        log.debug('Using provided custom steps between {} and {}'.format(
            steps[0], steps[-1]))

    # ------------------------------------------------------------------ #
    # Continue with trusted arguments
    # ------------------------------------------------------------------ #

    numsteps  = len(steps)        # number of steps for rks
    numtrials = data.shape[0]     # number of trials
    numels    = data.shape[1]     # number of measurements per trial

    if (ut._log_locals):
        log.debug('Trusted Locals: {}'.format(locals()))

    log.info("coefficients() with '{}' method for {} trials of length {}.{}" \
        .format(method, numtrials, numels,
            " 'knownmean' provided: {}".format(knownmean) \
            if knownmean is not None else ""))

    trialcrs        = []
    bootstrapcrs    = []
    stderrs         = None
    trialactivities = np.mean(data, axis=1, dtype=ftype)
    trialvariances  = np.var(data, axis=1, ddof=1, dtype=ftype)
    coefficients    = None                    # set later

    if method == 'trialseparated':
        ts_prepped   = ts_precompute(data, steps, knownmean)
        coefficients = ts_method(ts_prepped, steps)

        # save per-trial result
        for tdx in range(numtrials):
            tempdesc = 'Trial {}'.format(tdx)
            if description is not None:
                tempdesc = '{} ({})'.format(description, tempdesc)
            temp = CoefficientResult(
                coefficients    = ts_prepped[tdx],
                trialactivities = np.array([trialactivities[tdx]]),
                trialvariances  = np.array([trialvariances[tdx]]),
                triallen        = numels,
                steps           = steps,
                dt              = dt,
                dtunit          = dtunit,
                description     = tempdesc)
            trialcrs.append(temp)

    elif method == 'stationarymean':
        sm_prepped   = sm_precompute(data, steps, knownmean)
        coefficients = sm_method(sm_prepped, steps)


    # ------------------------------------------------------------------ #
    # Bootstrapping
    # ------------------------------------------------------------------ #

    if numboot <= 1:
        log.debug('Bootstrap needs at least numboot=2 replicas, ' +
            'skipping the resampling')
    elif numtrials < 2:
        log.info('Bootstrapping needs at least 2 trials, skipping ' +
            'the resampling')
    elif numboot > 1:
        log.info('Bootstrapping {} replicas'.format(numboot))
        log.debug('coefficients() seeding to {}'.format(seed))
        if seed is None:
            pass
        elif seed == 'random':
            np.random.seed(None)
        else:
            np.random.seed(seed)

        bscoefficients = np.zeros(shape=(numboot, numsteps), dtype='float64')

        for tdx in range(numboot):
            # log.info('{}/{} replicas'.format(tdx+1, numboot))
            choices = np.random.choice(np.arange(0, numtrials), size=numtrials)
            bsmean  = np.mean(trialactivities[choices],         dtype=ftype)
            bsvar   = np.var( trialactivities[choices], ddof=1, dtype=ftype)

            if method == 'trialseparated':
                bscoefficients[tdx] = ts_method(ts_prepped, steps, choices)

            elif method == 'stationarymean':
                bscoefficients[tdx] = sm_method(sm_prepped, steps, choices)

            tempdesc = 'Bootstrap Replica {}'.format(tdx)
            if description is not None:
                tempdesc = '{} ({})'.format(description, tempdesc)
            temp = CoefficientResult(
                coefficients    = bscoefficients[tdx],
                trialactivities = np.array([bsmean]),
                trialvariances  = np.array([bsvar]),
                triallen        = numels,
                steps           = steps,
                dt              = dt,
                dtunit          = dtunit,
                description     = tempdesc)
            bootstrapcrs.append(temp)

        log.info('{} bootstrap replicas done'.format(numboot))

        stderrs = np.sqrt(np.var(bscoefficients, axis=0, ddof=1, dtype=ftype))
        if (stderrs == stderrs[0]).all():
            stderrs = None

    fulres =  CoefficientResult(
        coefficients    = coefficients,
        trialactivities = trialactivities,
        trialvariances  = trialvariances,
        triallen        = numels,
        steps           = steps,
        stderrs         = stderrs,
        trialcrs        = trialcrs,
        bootstrapcrs    = bootstrapcrs,
        dt              = dt,
        dtunit          = dtunit,
        method          = method,
        description     = description)

    return fulres
